/* solution to problem C06 from the "many digits friendly competition":
   compute the first 10^N decimal digits after the decimal point of
   arctanh(1-arctanh(1-arctanh(1-arctanh(1/Pi)))) ~ 1.123.

   We compute with precision p and rounding to nearest:

   q = o(Pi)
   r = o(1/q)
   s = o(arctahnh(r))
   t = o(1-s)
   u = o(arctahnh(t))
   v = o(1-u)
   w = o(arctahnh(v))
   x = o(1-w)
   y = o(arctahnh(x)).

   We can check that for p >= 5, we have 2 <= q < 4, 1/4 <= r, s < 1/2,
   1/2 <= t, u, x < 1 [t <= 11/16, x <= 27/32], 1/8 <= v, w < 1/4, 1 <= y < 2.

   p = 6;
   q = [approx(Pi,p,3),approx(Pi,p,2)]
   r = [approx(1/q[2],p,3),approx(1/q[1],p,2)]
   s = [approx(atanh(r[1]),p,3),approx(atanh(r[2]),p,2)]
   t = [approx(1-s[2],p,3),approx(1-s[1],p,2)]
   u = [approx(atanh(t[1]),p,3),approx(atanh(t[2]),p,2)]
   v = [approx(1-u[2],p,3),approx(1-u[1],p,2)]
   w = [approx(atanh(v[1]),p,3),approx(atanh(v[2]),p,2)]
   x = [approx(1-w[2],p,3),approx(1-w[1],p,2)]
   y = [approx(atanh(x[1]),p,3),approx(atanh(x[2]),p,2)]

   We have q = Pi * (1+theta) with |theta| <= 2^(-p),
   thus r = 1/Pi * (1+theta)^2, which can be written 1/Pi * (1+3*theta),
   thus the absolute error on r is at most 2^(-p).

   The error on s is at most 1/2*ulp(s) + err(r) / (1-alpha^2)
   <= 2^(-p-2) + 2^(-p) / (1-(1/2)^2) <= 2 * 2^(-p).

   The error on t is at most 1/2*ulp(t) + err(s) <= 2^(-p-1) + 2 * 2^(-p)
   <= 3 * 2^(-p).

   The error on u is at most 1/2*ulp(u) + err(t) / (1-alpha^2)
   <= 2^(-p-1) + 3 * 2^(-p) / (1-(11/16)^2) <= 7 * 2^(-p).

   The error on v is at most 1/2*ulp(v) + err(u) <= 2^(-p-3) + 7 * 2^(-p)
   <= 8 * 2^(-p).

   The error on w is at most 1/2*ulp(w) + err(v) / (1-alpha^2)
   <= 2^(-p-3) + 8 * 2^(-p) / (1 - (1/4)^2) <= 9 * 2^(-p).

   The error on x is at most 1/2*ulp(x) + err(w) <= 10 * 2^(-p).

   Finally, the error on y is at most 1/2*ulp(y) + err(x) / (1-alpha^2)
   <= 2^(-p) + 10 * 2^(-p) / (1 - (27/32)^2) <= 36 * 2^(-p) <= 2^(6-p).
*/

#include "many.h"

int
main (int argc, char *argv[])
{
  unsigned long N = atoi (argv[1]), M;
  mp_prec_t p;
  mpfr_t i;
  char *lo;
  mp_exp_t exp_lo;
  int st, st0;

  fprintf (stderr, "Using GMP %s and MPFR %s\n", gmp_version, mpfr_version);
  st = cputime ();

  mpfr_init (i);

  M = N + 1;

  do
    {
      M += 10;
      mpfr_set_prec (i, 32);
      mpfr_set_d (i, LOG2_10, GMP_RNDU);
      mpfr_mul_ui (i, i, M, GMP_RNDU);
      mpfr_add_ui (i, i, 6, GMP_RNDU);
      p = mpfr_get_ui (i, GMP_RNDU);
      fprintf (stderr, "Setting precision to %lu\n", p);

      mpfr_set_prec (i, p);

      mpfr_const_pi (i, GMP_RNDN);
      mpfr_ui_div (i, 1, i, GMP_RNDN);
      mpfr_atanh (i, i, GMP_RNDN);
      mpfr_ui_sub (i, 1, i, GMP_RNDN);
      mpfr_atanh (i, i, GMP_RNDN);
      mpfr_ui_sub (i, 1, i, GMP_RNDN);
      mpfr_atanh (i, i, GMP_RNDN);
      mpfr_ui_sub (i, 1, i, GMP_RNDN);
      mpfr_atanh (i, i, GMP_RNDN);

      st0 = cputime ();
      lo = mpfr_get_str (NULL, &exp_lo, 10, M, i, GMP_RNDN);
      st0 = cputime () - st0;
    }
  while (can_round (lo, N + 1, M) == 0);

  lo ++;
  lo[N] = '\0';
  printf ("%s\n", lo);

  mpfr_clear (i);

  fprintf (stderr, "Cputime: %dms (output %dms)\n", cputime () - st, st0);
  return 0;
}
