/* solution to problem C09 from the "many digits friendly competition":
   compute the first 10^N decimal digits after the decimal point of
   sin(10*arctan(tanh(Pi*sqrt(2011)/3))) ~ 0.999.

   We compute with precision p and rounding to nearest:

   q = o(Pi)
   r = o(sqrt(2011))
   s = o(q*r)
   t = o(s/3)
   u = o(tanh(t))
   v = o(arctan(u))
   w = o(10*v)
   x = o(sin(w)).

   For p >= 6, we have 2 <= q < 4, 32 <= r, t < 64, 128 <= s < 256,
   1/2 <= u, v < 1, 4 <= w < 8.
   More precisely, 45 <= t <= 48, 63/64 <= u <= 1, 61/8 <= w <= 8,
   and sin(61/8) ~ 0.973 <= x <= 1.

   p=6;
   q = [approx(Pi,p,3),approx(Pi,p,2)]
   r = [approx(sqrt(2011),p,3),approx(sqrt(2011),p,2)]
   s = [approx(q[1]*r[1],p,3),approx(q[2]*r[2],p,2)]
   t = [approx(s[1]/3,p,3),approx(s[2]/3,p,2)]
   u = [approx(tanh(t[1]),p,3),approx(tanh(t[1]),p,2)]
   v = [approx(atan(u[1]),p,3),approx(atan(u[2]),p,2)]
   w = [approx(10*v[1],p,3),approx(10*v[2],p,2)]

   We have q = Pi * (1+theta) with |theta| <= 2^(-p),
   r = sqrt(2011) * (1+theta)
   s = Pi*sqrt(2011) * (1+theta)^3
   t = Pi*sqrt(2011)/3 * (1+theta)^4

   Since (1+theta)^4 can be written 1+5*theta, the absolute error on t
   is bounded by Pi*sqrt(2011)/3 * (5*theta) <= 235 * 2^(-p).
   Thus err(u) <= 1/2*ulp(u) + err(t) * (1-tanh(alpha)^2)
               <= 2^(-p-1) + 235 * 2^(-p) * (1-tanh(45)^2)
	       <= 2^(-p-1) + 2^(-119-p)
	       <= 2^(-p).

   err(v) <= 1/2*ulp(v) + err(u) / (1 + alpha^2)
          <= 2^(-p-1) + 2^(-p) / (1 + (63/64)^2)
	  <= 2 * 2^(-p).

   err(w) <= 1/2*ulp(w) + 10*err(v)
          <= 2^(2-p) + 20*2^(-p)
	  <= 24 * 2^(-p).

   Finally, err(x) <= 1/2*ulp(x) + err(w) * cos(alpha)
                   <= 2^(-p-1) + 24 * 2^(-p)
		   <= 2^(5-p).
*/

#include "many.h"

int
main (int argc, char *argv[])
{
  unsigned long N, M;
  mp_prec_t p;
  mpfr_t i, j, k;
  char *lo;
  mp_exp_t exp_lo;
  int st, st0 = 0, verbose = 0;

  if (argc >= 2 && strcmp (argv[1], "-v") == 0)
    {
      verbose = 1;
      argv ++;
      argc --;
    }

  N = atoi (argv[1]);

  fprintf (stderr, "Using GMP %s and MPFR %s\n", gmp_version, mpfr_version);
  st = cputime ();

  mpfr_init (i);
  mpfr_init (j);
  mpfr_init (k);

  M = N;

  do
    {
      M += 10;
      mpfr_set_prec (i, 32);
      mpfr_set_d (i, LOG2_10, GMP_RNDU);
      mpfr_mul_ui (i, i, M, GMP_RNDU);
      mpfr_add_ui (i, i, 6, GMP_RNDU);
      p = mpfr_get_ui (i, GMP_RNDU);
      if (p < 11)
	p = 11; /* so that 2011 is exact */
      if (verbose)
        fprintf (stderr, "Setting precision to %lu\n", p);

      mpfr_set_prec (i, p);
      mpfr_set_prec (j, p);

      st0 = cputime ();
      mpfr_const_pi (i, GMP_RNDN);
      if (verbose)
        fprintf (stderr, "mpfr_const_pi() took %dms\n", cputime () - st0);
      mpfr_set_ui (j, 2011, GMP_RNDN);
      mpfr_sqrt (j, j, GMP_RNDN);
      mpfr_mul (i, i, j, GMP_RNDN);
      mpfr_div_ui (i, i, 3, GMP_RNDN);
      st0 = cputime ();
      mpfr_tanh (i, i, GMP_RNDN);
      if (verbose)
        fprintf (stderr, "mpfr_tanh() took %dms\n", cputime () - st0);
      st0 = cputime ();
      mpfr_atan (i, i, GMP_RNDN);
      if (verbose)
        fprintf (stderr, "mpfr_atan() took %dms\n", cputime () - st0);
      mpfr_mul_ui (i, i, 10, GMP_RNDN);
      st0 = cputime ();
      mpfr_sin (i, i, GMP_RNDN);
      if (verbose)
        fprintf (stderr, "mpfr_sin() took %dms\n", cputime () - st0);

      st0 = cputime ();
      lo = mpfr_get_str (NULL, &exp_lo, 10, M, i, GMP_RNDN);
      if (verbose)
        fprintf (stderr, "mpfr_get_str() took %dms\n", cputime () - st0);
    }
  while (can_round (lo, N, M) == 0);

  lo[N] = '\0';
  printf ("%s\n", lo);

  mpfr_clear (i);
  mpfr_clear (j);
  mpfr_clear (k);

  fprintf (stderr, "Cputime: %dms\n", cputime () - st);
  return 0;
}
