/* solution to problem C15 from the "many digits friendly competition":
   compute the first 10^N decimal digits after the decimal point of
   h(10^(N+1)) where
   h(n) = sum(1/k, k=n..n^2).

   According to (6.3.2) in Abramowitz & Stegun, we have:

   Psi(n) = -gamma + sum(1/k, k=1..n-1)

   thus h(n) = Psi(n^2+1) - Psi(n) and we are reduced to the problem of
   computing n/10 digits of Psi(n^2+1) - Psi(n).
   
   Formula (6.3.18) from Abramowitz & Stegun gives:

   Psi(z) ~ log(z) - 1/(2z) - sum(B[2k]/(2kz^(2k)), k=1..infinity).

   Since we have |B[2k]| <= 2*(2k)!/(2Pi)^(2k)/(1-2^(1-2k)) (23.1.15),
   we deduce |B[2k]/(2kz^(2k))| <= 2/k (2k)!/(2Piz)^(2k) for k >= 1

   Using n! <= (n/e)^n*sqrt(8*n) which is true for n >= 1, we get

   |B[2k]/(2kz^(2k))| <= 8 * (k/(ePiz))^(2k).

   The minimum is of (k/(ePiz))^(2k) is obtained for k ~ Piz,
   with a value of e^(-2k) ~ e^(-2Piz). Thus to get a sufficient
   accuracy, we need that z is large enough.

   Here we have:

   (a) for Psi(n): z=n and we want n/10 digits, i.e. an error of 10^(-n/10).
       We have e^(-2Piz) ~ 2^(-9.06n) and 10^(-n/10) ~ 2^(-0.33n), thus ok.
   (b) for Psi(n^2+1): z=n^2+1 and we still want n/10 digits. This is ok too.

   Let R(z) = sum(B[2k]/(2kz^(2k)), k=1..infinity).
   
   We have Psi(n^2+1) - Psi(n) ~ log((n^2+1)/n) - 1/(2(n^2+1)) + 1/(2n)
                                 + R(n) - R(n^2+1).
*/

#include "many.h"

/* [Copied from mpfr/lngamma.c.]
   assuming b[0]...b[2(n-1)] are computed, computes and stores B[2n]*(2n+1)!

   t/(exp(t)-1) = sum(B[j]*t^j/j!, j=0..infinity)
   thus t = (exp(t)-1) * sum(B[j]*t^j/j!, n=0..infinity).
   Taking the coefficient of degree n+1 > 1, we get:
   0 = sum(1/(n+1-k)!*B[k]/k!, k=0..n)
   which gives:
   B[n] = -sum(binomial(n+1,k)*B[k], k=0..n-1)/(n+1).

   Let C[n] = B[n]*(n+1)!.
   Then C[n] = -sum(binomial(n+1,k)*C[k]*n!/(k+1)!,  k=0..n-1),
   which proves that the C[n] are integers.
*/
static mpz_t*
bernoulli (mpz_t *b, unsigned long n)
{
  if (n == 0)
    {
      b = (mpz_t *) (*__gmp_allocate_func) (sizeof (mpz_t));
      mpz_init_set_ui (b[0], 1);
    }
  else
    {
      mpz_t t;
      unsigned long k;

      b = (mpz_t *) (*__gmp_reallocate_func)
        (b, n * sizeof (mpz_t), (n + 1) * sizeof (mpz_t));
      mpz_init (b[n]);
      /* b[n] = -sum(binomial(2n+1,2k)*C[k]*(2n)!/(2k+1)!,  k=0..n-1) */
      mpz_init_set_ui (t, 2 * n + 1);
      mpz_mul_ui (t, t, 2 * n - 1);
      mpz_mul_ui (t, t, 2 * n);
      mpz_mul_ui (t, t, n);
      mpz_div_ui (t, t, 3); /* exact: t=binomial(2*n+1,2*k)*(2*n)!/(2*k+1)!
                               for k=n-1 */
      mpz_mul (b[n], t, b[n-1]);
      for (k = n - 1; k-- > 0;)
        {
	  if (k < 128)
	    mpz_mul_ui (t, t, (2 * k + 1) * (2 * k + 2) * (2 * k + 2) * (2 * k + 3));
	  else if (k < 32767)
	    {
	      mpz_mul_ui (t, t, (2 * k + 1) * (2 * k + 3));
	      mpz_mul_ui (t, t, (2 * k + 2) * (2 * k + 2));
	    }
	  else
	    {
	      mpz_mul_ui (t, t, 2 * k + 1);
	      mpz_mul_ui (t, t, 2 * k + 2);
	      mpz_mul_ui (t, t, 2 * k + 2);
	      mpz_mul_ui (t, t, 2 * k + 3);
	    }
	  if (n - k < 32768)
	    mpz_div_ui (t, t, (2 * (n - k) + 1) * 2 * (n - k));
	  else
	    {
	      mpz_div_ui (t, t, 2 * (n - k) + 1);
	      mpz_div_ui (t, t, 2 * (n - k));
	    }
          mpz_addmul (b[n], t, b[k]);
        }
      /* take into account C[1] */
      mpz_mul_ui (t, t, 2 * n + 1);
      mpz_div_2exp (t, t, 1);
      mpz_sub (b[n], b[n], t);
      mpz_neg (b[n], b[n]);
      mpz_clear (t);
    }
  return b;
}

int
main (int argc, char *argv[])
{
  unsigned long N = atoi (argv[1]), M, k, n;
  mp_prec_t p;
  mpfr_t i, j, t, s;
  char *lo;
  int st, st0, stb = 0;
  mp_exp_t e;
  mpz_t *B, tmp;
  unsigned long Bm = 0; /* number of allocated B[] */
  unsigned long oldBm;
  int overflow;

  fprintf (stderr, "Using GMP %s and MPFR %s\n", gmp_version, mpfr_version);
  st = cputime ();

  mpfr_init (i);
  mpfr_init (j);
  mpfr_init (t);
  mpfr_init (s);
  mpz_init (tmp);

  n = 10 * N;

  /* h(10*N) is about log(10*N).
     For N<=10^7, this is less than 8 digits. */
  M = N + 8;

  do
    {
      M += 10;
      mpfr_set_prec (i, 32);
      mpfr_set_d (i, LOG2_10, GMP_RNDU);
      mpfr_mul_ui (i, i, M, GMP_RNDU);
      mpfr_add_ui (i, i, 6, GMP_RNDU);
      p = mpfr_get_ui (i, GMP_RNDU);
      fprintf (stderr, "Setting precision to %lu\n", p);

      mpfr_set_prec (i, p);
      mpfr_set_prec (j, p);
      mpfr_set_prec (t, p);
      mpfr_set_prec (s, p);

      if (Bm == 0)
        {
          B = bernoulli ((mpz_t *) 0, 0);
          B = bernoulli (B, 1);
          Bm = 2;
        }

      /* first compute R(n) in s */
      mpfr_set_ui (s, 0, GMP_RNDN); /* sum */
      mpfr_set_ui (j, 1, GMP_RNDN); /* 1/(2k+1)!/z^(2k) */
      for (k = 1; ; k++)
	{
	  mpfr_div_ui (j, j, 2 * k, GMP_RNDN);
	  mpfr_div_ui (j, j, 2 * k + 1, GMP_RNDN);
	  mpfr_div_ui (j, j, n, GMP_RNDN);
	  mpfr_div_ui (j, j, n, GMP_RNDN);
	  /* compute C[2k] */
          if (Bm <= k)
            {
	      stb -= cputime ();
              B = bernoulli (B, k); /* B[2m]*(2m+1)!, exact */
	      stb += cputime ();
              Bm ++;
            }
	  mpfr_mul_z (t, j, B[k], GMP_RNDN);
	  mpfr_div_ui (t, t, 2 * k, GMP_RNDN);
	  mpfr_add (s, s, t, GMP_RNDN);
	  if (mpfr_get_exp (t) + (mp_exp_t) p < 0)
	    break;
	}
      fprintf (stderr, "Psi(n) needed %lu terms\n", k);
      fprintf (stderr, "last B[2k] has %lu bits\n", mpz_sizeinbase (B[k], 2));
      fprintf (stderr, "bernoulli took %dms\n", stb);

      /* now compute R(n^2+1) and subtract it to s */
      mpfr_set_ui (j, 1, GMP_RNDN); /* 1/(2k+1)!/z^(2k) */
      overflow = n >= 1 << (mp_bits_per_limb / 2);
      if (overflow)
	{
	  mpz_set_ui (tmp, n);
	  mpz_mul (tmp, tmp, tmp); /* n^2 */
	  mpz_add_ui (tmp, tmp, 1); /* n^2+1 */
	  mpz_mul (tmp, tmp, tmp); /* (n^2+1)^2 */
	  mpfr_set_prec (i, mpz_sizeinbase (tmp, 2));
	  mpfr_set_z (i, tmp, GMP_RNDN);
	}
      for (k = 1; ; k++)
	{
	  mpfr_div_ui (j, j, 2 * k, GMP_RNDN);
	  mpfr_div_ui (j, j, 2 * k + 1, GMP_RNDN);
	  if (overflow)
	    {
	      mpfr_div (j, j, i, GMP_RNDN);
	    }
	  else
	    {
	      mpfr_div_ui (j, j, n * n + 1, GMP_RNDN);
	      mpfr_div_ui (j, j, n * n + 1, GMP_RNDN);
	    }
	  /* compute C[2k] */
          if (Bm <= k) /* should not happen here */
            {
	      abort ();
              B = bernoulli (B, k); /* B[2m]*(2m+1)!, exact */
              Bm ++;
            }
	  mpfr_mul_z (t, j, B[k], GMP_RNDN);
	  mpfr_div_ui (t, t, 2 * k, GMP_RNDN);
	  mpfr_sub (s, s, t, GMP_RNDN);
	  if (mpfr_get_exp (t) + (mp_exp_t) p < 0)
	    break;
	}
      fprintf (stderr, "Psi(n^2+1) needed %lu terms\n", k);
      if (overflow)
	mpfr_set_prec (i, p);

      /* add log((n^2+1)/n) */
      mpz_set_ui (tmp, n);
      mpz_mul_ui (tmp, tmp, n);
      mpz_add_ui (tmp, tmp, 1); /* n^2+1 */
      mpfr_set_z (i, tmp, GMP_RNDN);
      mpfr_div_ui (i, i, n, GMP_RNDN);
      mpfr_log (i, i, GMP_RNDN);
      mpfr_add (s, s, i, GMP_RNDN);

      /* add 1/(2n) - 1/(2(n^2+1)) = 1/2 * (n^2 - n + 1)/(n^2 + 1)/n */
      mpfr_set_ui (i, n * n - n + 1, GMP_RNDN);
      mpfr_set_z (i, tmp, GMP_RNDN); /* n^2+1 */
      mpfr_sub_ui (i, i, n, GMP_RNDN); /* n^2 - n + 1 */
      mpfr_div_z (i, i, tmp, GMP_RNDN);
      mpfr_div_ui (i, i, n, GMP_RNDN);
      mpfr_div_2exp (i, i, 1, GMP_RNDN);
      mpfr_add (s, s, i, GMP_RNDN);

      st0 = cputime ();
      lo = mpfr_get_str (NULL, &e, 10, M, s, GMP_RNDN);
      st0 = cputime () - st0;
    }
  while (can_round (lo + e, N, M - e) == 0);

  lo += e;
  lo[N] = '\0';
  printf ("%s\n", lo);

  oldBm = Bm;
  while (Bm--)
    mpz_clear (B[Bm]);
  (*__gmp_free_func) (B, oldBm * sizeof (mpz_t));

  mpfr_clear (i);
  mpfr_clear (j);
  mpfr_clear (t);
  mpfr_clear (s);
  mpz_clear (tmp);

  fprintf (stderr, "Cputime: %dms (output %dms)\n", cputime () - st, st0);
  return 0;
}
