#include <assert.h>
#include "many.h"

#ifndef MPFR_PREC
#define MPFR_PREC(z) mpfr_get_prec(z)
#endif

#ifndef MPFR_INT_CEIL_LOG2
#define MPFR_INT_CEIL_LOG2(n) (unsigned long) ceil(log((double) n) * 1.4426950408889634)
#endif

int
zeta_ui (mpfr_ptr z, unsigned long m, mp_rnd_t r)
{
  mp_prec_t p = MPFR_PREC(z);
  unsigned long n, k, err, kbits;
  mpz_t d, t, s, q;
  mpfr_t y;
  int inex;
  int loop = 0;

  assert((m == 2 || m == 3 || m == 5) && m < p);

  mpz_init (s);
  mpz_init (d);
  mpz_init (t);
  mpz_init (q);
  mpfr_init2 (y, MPFR_PREC_MIN);

  p += MPFR_INT_CEIL_LOG2(p); /* account of the n term in the error */

  do
    {
      if (loop++)
        fprintf (stderr, "recompute\n");
      p += MPFR_INT_CEIL_LOG2(p) + 1;
      /* 0.39321985067869744 = log(2)/log(3+sqrt(8)) */
      n = 1 + (unsigned long) (0.39321985067869744 * (double) p);
      err = n + 4;

      mpfr_set_prec (y, p);

      /* computation of the d[k] */
      mpz_set_ui (s, 0);
      mpz_set_ui (t, 1);
      mpz_mul_2exp (t, t, 2 * n - 1); /* t[n] */
      mpz_set (d, t);
      for (k = n; k > 0; k--)
        {
          mpz_ui_pow_ui (q, k, m);
          kbits = mpz_sizeinbase (q, 2);
          /* if k^m is too large, use mpz_tdiv_q */
          if (kbits > 2 * GMP_LIMB_BITS)
            {
              mpz_tdiv_q (q, d, q);
            }
          else /* use several mpz_tdiv_q_ui calls */
            {
              unsigned long km = k, mm = m - 1;
              while (mm > 0 && km < ULONG_MAX / k)
                {
                  km *= k;
                  mm --;
                }
              mpz_tdiv_q_ui (q, d, km);
              while (mm > 0)
                {
                  km = k;
                  mm --;
                  while (mm > 0 && km < ULONG_MAX / k)
                    {
                      km *= k;
                      mm --;
                    }
                  mpz_tdiv_q_ui (q, q, km);
                }
            }
          if (k % 2)
            mpz_add (s, s, q);
          else
            mpz_sub (s, s, q);

          /* we have d[k] = sum(t[i], i=k+1..n)
             with t[i] = n*(n+i-1)!*4^i/(n-i)!/(2i)!
             t[k-1]/t[k] = (2k)*(2k-1)/(n-k+1)/(n+k-1)/4 */
#if (GMP_LIMB_BITS == 32)
#define KMAX 46341 /* max k such that k*(2k-1) < 2^32 */
#elif (GMP_LIMB_BITS == 64)
#define KMAX 3037000500
#endif
#ifdef KMAX
          if (k <= KMAX)
            mpz_mul_ui (t, t, k * (2 * k - 1));
          else
#endif
            {
              mpz_mul_ui (t, t, k);
              mpz_mul_ui (t, t, 2 * k - 1);
            }
          mpz_div_2exp (t, t, 1);
          if (n < 1UL << (GMP_LIMB_BITS / 2))
            {
              /* (n - k + 1) * (n + k - 1) < n^2 */
              mpz_divexact_ui (t, t, (n - k + 1) * (n + k - 1));
            }
          else
            {
              mpz_divexact_ui (t, t, n - k + 1);
              mpz_divexact_ui (t, t, n + k - 1);
            }
          mpz_add (d, d, t);
        }

      /* multiply by 1/(1-2^(1-m)) = 1 + 2^(1-m) + 2^(2-m) + ... */
      mpz_div_2exp (t, s, m - 1);
      do
        {
          err ++;
          mpz_add (s, s, t);
          mpz_div_2exp (t, t, m - 1);
        }
      while (mpz_cmp_ui (t, 0) > 0);

      /* divide by d[n] */
      mpz_mul_2exp (s, s, p);
      mpz_tdiv_q (s, s, d);
      mpfr_set_z (y, s, GMP_RNDN);
      mpfr_div_2exp (y, y, p, GMP_RNDN);

      err = MPFR_INT_CEIL_LOG2 (err);
    }
  while (mpfr_can_round (y, p - err, GMP_RNDN, GMP_RNDZ,
                         MPFR_PREC(z) + (r == GMP_RNDN)) == 0);

  mpz_clear (d);
  mpz_clear (t);
  mpz_clear (q);
  mpz_clear (s);
  inex = mpfr_set (z, y, r);
  mpfr_clear (y);
  return inex;
}

int
main (int argc, char *argv[])
{
  unsigned long N = atoi (argv[1]), M;
  mp_prec_t p;
  mpfr_t i, j;
  char *lo;
  mp_exp_t exp_lo;
  int st, st0;

  fprintf (stderr, "Using GMP %s and MPFR %s\n", gmp_version, mpfr_version);
  st = cputime ();

  mpfr_init (i);
  mpfr_init (j);

  M = N;

  do
    {
      M += 10;
      mpfr_set_prec (i, 32);
      mpfr_set_d (i, LOG2_10, GMP_RNDU);
      mpfr_mul_ui (i, i, M, GMP_RNDU);
      mpfr_add_ui (i, i, 4, GMP_RNDU);
      p = mpfr_get_ui (i, GMP_RNDU);
      fprintf (stderr, "Setting precision to %lu\n", p);

      mpfr_set_prec (i, p);
      mpfr_set_prec (j, p);

      zeta_ui (i, 2, GMP_RNDN);
      mpfr_mul_2ui (i, i, 1, GMP_RNDN);
      mpfr_sub_ui (i, i, 1, GMP_RNDN);
      zeta_ui (j, 3, GMP_RNDN);
      mpfr_sub_ui (j, j, 1, GMP_RNDN);
      mpfr_mul (i, i, j, GMP_RNDN);
      zeta_ui (j, 5, GMP_RNDN);
      mpfr_sub_ui (j, j, 1, GMP_RNDN);
      mpfr_add (i, i, j, GMP_RNDN);
      mpfr_mul_2ui (i, i, 1, GMP_RNDN);
      /* Absolute error < 2 * (2*0.21 + 2.29 + 0.25 + 1 + 0.5) * 2^(-p)
                        < 9 * 2^(-p)
         We chose p to satisfy:
           p > log2(9 * 10^M) = log2(9) + M.log2(10) */

      st0 = cputime ();
      lo = mpfr_get_str (NULL, &exp_lo, 10, M, i, GMP_RNDN);
      st0 = cputime () - st0;
    }
  while (can_round (lo, N, M) == 0);

  lo[N] = '\0';
  printf ("%s\n", lo);

  mpfr_clear (i);
  mpfr_clear (j);

  fprintf (stderr, "Cputime: %dms (output %dms)\n", cputime () - st, st0);
  return 0;
}
