#include <assert.h>
#include "many.h"

#ifndef MPFR_PREC
#define MPFR_PREC(z) mpfr_get_prec(z)
#endif

#ifndef MPFR_INT_CEIL_LOG2
#define MPFR_INT_CEIL_LOG2(n) (unsigned long) ceil(log((double) n) * 1.4426950408889634)
#endif

static void
zeta_ui (mpfr_ptr z3, mpfr_ptr z5, mp_prec_t p)
{
  mp_prec_t pz;
  unsigned long n, k, err3, err5, kbits;
  mpz_t d, t, s3, s5, q3, q5;
  int loop = 0;

  assert(p > 5);
  pz = p;

  mpz_init (s3);
  mpz_init (s5);
  mpz_init (d);
  mpz_init (t);
  mpz_init (q3);
  mpz_init (q5);

  p += MPFR_INT_CEIL_LOG2(p); /* account of the n term in the error */

  do
    {
      if (loop++)
        fprintf (stderr, "recompute\n");
      p += MPFR_INT_CEIL_LOG2(p) + 1;
      /* 0.39321985067869744 = log(2)/log(3+sqrt(8)) */
      n = 1 + (unsigned long) (0.39321985067869744 * (double) p);
      err3 = err5 = n + 4;

      mpfr_set_prec (z3, p);
      mpfr_set_prec (z5, p);

      /* computation of the d[k] */
      mpz_set_ui (s3, 0);
      mpz_set_ui (s5, 0);
      mpz_set_ui (t, 1);
      mpz_mul_2exp (t, t, 2 * n - 1); /* t[n] */
      mpz_set (d, t);
      for (k = n; k > 0; k--)
        {
          mpz_ui_pow_ui (q3, k, 3);
          kbits = mpz_sizeinbase (q3, 2);
          /* if k^m is too large, use mpz_tdiv_q */
          if (kbits > 2 * GMP_LIMB_BITS)
            {
              mpz_tdiv_q (q3, d, q3);
            }
          else /* use several mpz_tdiv_q_ui calls */
            {
              unsigned long km = k, mm = 3 - 1;
              while (mm > 0 && km < ULONG_MAX / k)
                {
                  km *= k;
                  mm --;
                }
              mpz_tdiv_q_ui (q3, d, km);
              while (mm > 0)
                {
                  km = k;
                  mm --;
                  while (mm > 0 && km < ULONG_MAX / k)
                    {
                      km *= k;
                      mm --;
                    }
                  mpz_tdiv_q_ui (q3, q3, km);
                }
            }
          if (k % 2)
            mpz_add (s3, s3, q3);
          else
            mpz_sub (s3, s3, q3);

          mpz_ui_pow_ui (q5, k, 5);
          kbits = mpz_sizeinbase (q5, 2);
          /* if k^m is too large, use mpz_tdiv_q */
          if (kbits > 2 * GMP_LIMB_BITS)
            {
              mpz_tdiv_q (q5, d, q5);
            }
          else /* use several mpz_tdiv_q_ui calls */
            {
              unsigned long km = k, mm = 5 - 1;
              while (mm > 0 && km < ULONG_MAX / k)
                {
                  km *= k;
                  mm --;
                }
              mpz_tdiv_q_ui (q5, d, km);
              while (mm > 0)
                {
                  km = k;
                  mm --;
                  while (mm > 0 && km < ULONG_MAX / k)
                    {
                      km *= k;
                      mm --;
                    }
                  mpz_tdiv_q_ui (q5, q5, km);
                }
            }
          if (k % 2)
            mpz_add (s5, s5, q5);
          else
            mpz_sub (s5, s5, q5);

          /* we have d[k] = sum(t[i], i=k+1..n)
             with t[i] = n*(n+i-1)!*4^i/(n-i)!/(2i)!
             t[k-1]/t[k] = (2k)*(2k-1)/(n-k+1)/(n+k-1)/4 */
#if (GMP_LIMB_BITS == 32)
#define KMAX 46341 /* max k such that k*(2k-1) < 2^32 */
#elif (GMP_LIMB_BITS == 64)
#define KMAX 3037000500
#endif
#ifdef KMAX
          if (k <= KMAX)
            mpz_mul_ui (t, t, k * (2 * k - 1));
          else
#endif
            {
              mpz_mul_ui (t, t, k);
              mpz_mul_ui (t, t, 2 * k - 1);
            }
          mpz_div_2exp (t, t, 1);
          if (n < 1UL << (GMP_LIMB_BITS / 2))
            {
              /* (n - k + 1) * (n + k - 1) < n^2 */
              mpz_divexact_ui (t, t, (n - k + 1) * (n + k - 1));
            }
          else
            {
              mpz_divexact_ui (t, t, n - k + 1);
              mpz_divexact_ui (t, t, n + k - 1);
            }
          mpz_add (d, d, t);
        }

      /* multiply by 1/(1-2^(1-m)) = 1 + 2^(1-m) + 2^(2-m) + ... */
      mpz_div_2exp (t, s3, 3 - 1);
      do
        {
          err3 ++;
          mpz_add (s3, s3, t);
          mpz_div_2exp (t, t, 3 - 1);
        }
      while (mpz_cmp_ui (t, 0) > 0);

      /* divide by d[n] */
      mpz_mul_2exp (s3, s3, p);
      mpz_tdiv_q (s3, s3, d);

      mpfr_set_z (z3, s3, GMP_RNDN);
      mpfr_div_2exp (z3, z3, p, GMP_RNDN);

      /* multiply by 1/(1-2^(1-m)) = 1 + 2^(1-m) + 2^(2-m) + ... */
      mpz_div_2exp (t, s5, 5 - 1);
      do
        {
          err5 ++;
          mpz_add (s5, s5, t);
          mpz_div_2exp (t, t, 5 - 1);
        }
      while (mpz_cmp_ui (t, 0) > 0);

      /* divide by d[n] */
      mpz_mul_2exp (s5, s5, p);
      mpz_tdiv_q (s5, s5, d);

      mpfr_set_z (z5, s5, GMP_RNDN);
      mpfr_div_2exp (z5, z5, p, GMP_RNDN);

      err3 = MPFR_INT_CEIL_LOG2 (err3);
      err5 = MPFR_INT_CEIL_LOG2 (err5);
    }
  while (mpfr_can_round (z3, p - err3, GMP_RNDN, GMP_RNDZ, pz + 1) == 0 &&
         mpfr_can_round (z5, p - err5, GMP_RNDN, GMP_RNDZ, pz + 1) == 0);

  mpz_clear (d);
  mpz_clear (t);
  mpz_clear (q3);
  mpz_clear (q5);
  mpz_clear (s3);
  mpz_clear (s5);
  mpfr_prec_round (z3, pz, GMP_RNDN);
  mpfr_prec_round (z5, pz, GMP_RNDN);
}

int
main (int argc, char *argv[])
{
  unsigned long N = atoi (argv[1]), M;
  mp_prec_t p;
  mpfr_t i, j, k;
  char *lo;
  mp_exp_t exp_lo;
  int st, st0;

  fprintf (stderr, "Using GMP %s and MPFR %s\n", gmp_version, mpfr_version);
  st = cputime ();

  mpfr_init (i);
  mpfr_init (j);
  mpfr_init (k);

  M = N;

  do
    {
      M += 10;
      mpfr_set_prec (i, 32);
      mpfr_set_d (i, LOG2_10, GMP_RNDU);
      mpfr_mul_ui (i, i, M, GMP_RNDU);
      mpfr_add_ui (i, i, 4, GMP_RNDU);
      p = mpfr_get_ui (i, GMP_RNDU);
      fprintf (stderr, "Setting precision to %lu\n", p);

      mpfr_set_prec (i, p);
      mpfr_set_prec (j, p);
      mpfr_set_prec (k, p);

      mpfr_const_pi (i, GMP_RNDN); /* err <= 2^(-p+1) */
      mpfr_sqr (i, i, GMP_RNDN); /* err < 21 * 2^(-p) */
      mpfr_div_ui (i, i, 3, GMP_RNDN); /* err < 9 * 2^(-p) */
      mpfr_sub_ui (i, i, 1, GMP_RNDN); /* err < 9 * 2^(-p) */
      zeta_ui (j, k, p); /* err <= 2^(-p) on j and k */
      mpfr_sub_ui (j, j, 1, GMP_RNDN); /* err <= 2^(-p) */
      mpfr_sub_ui (k, k, 1, GMP_RNDN); /* err <= 2^(-p) */
      mpfr_mul (i, i, j, GMP_RNDN); /* err < 4.5 * 2^(-p) */
      mpfr_add (i, i, k, GMP_RNDN); /* err < 6 * 2^(-p) */
      mpfr_mul_2ui (i, i, 1, GMP_RNDN); /* err < 12 * 2^(-p) */
      /* We chose p to satisfy:
           p > log2(12 * 10^M) = log2(12) + M.log2(10) */

      st0 = cputime ();
      lo = mpfr_get_str (NULL, &exp_lo, 10, M, i, GMP_RNDN);
      st0 = cputime () - st0;
    }
  while (can_round (lo, N, M) == 0);

  lo[N] = '\0';
  printf ("%s\n", lo);

  mpfr_clear (i);
  mpfr_clear (j);

  fprintf (stderr, "Cputime: %dms (output %dms)\n", cputime () - st, st0);
  return 0;
}
