/* implement Lehmer's method (page 4, equation (3)) from
   http://web.comlab.ox.ac.uk/oucl/work/richard.brent/pub/pub166.html
*/

#define USE_MPFI
#include "many.h"

/* alpha is an interval enclosing the true value */
static unsigned long *
confrac (mpfi_t alpha, unsigned long *n)
{
  unsigned long *l, alloc, a;
  mpz_t beta0, beta1, gamma0, gamma1, g0, g1;
  mp_prec_t prec = mpfi_get_prec (alpha);
  mp_size_t n0;

  alloc = 1;
  l = malloc (alloc * sizeof (unsigned long));
  *n = 0; /* number of stored elements */

  mpz_init (beta0);
  mpz_init (beta1);
  mpz_init (gamma0);
  mpz_init (gamma1);
  mpz_init (g0);
  mpz_init (g1);

  mpfi_mul_2exp (alpha, alpha, prec);
  mpfr_get_z (beta1, RIGHT(alpha), GMP_RNDU);
  mpfr_get_z (gamma1, LEFT(alpha), GMP_RNDD);
  mpz_set_ui (beta0, 1);
  mpz_mul_2exp (beta0, beta0, prec);
  mpz_set (gamma0, beta0);

  /* beta0/beta1 < 1/alpha < gamma0/gamma1 */

  while (1)
    {
      /* we must have gamma0 >= gamma1 */
      n0 = mpz_sizeinbase (gamma0, 2);
      n0 = (n0 >= mp_bits_per_limb) ? n0 - mp_bits_per_limb : 0;
      mpz_tdiv_q_2exp (g0, gamma0, n0);
      mpz_cdiv_q_2exp (g1, gamma1, n0);

      /* we have g0 = floor(gamma0/2^n0) and g1 = ceil(gamma1/2^n0)
         thus floor (g0/g1) <= floor(gamma0/gamma1) */

      a = mpz_get_ui (g0) / mpz_get_ui (g1);

      mpz_submul_ui (gamma0, gamma1, a);
      /* if a was too small, then gamma0 >= gamma1 */
      while (mpz_cmp (gamma0, gamma1) > 0)
        {
          mpz_sub (gamma0, gamma0, gamma1);
          a ++;
        }
      mpz_submul_ui (beta0, beta1, a);
      if (mpz_cmp_ui (beta0, 0) < 0 || mpz_cmp (beta0, beta1) >= 0)
        break;
      mpz_swap (gamma0, gamma1);
      mpz_swap (beta0, beta1);

      if (*n >= alloc)
        {
          alloc = 2 * alloc;
          l = realloc (l, alloc * sizeof (unsigned long));
        }
      l[*n] = a;
      (*n)++;

    }

  mpz_clear (beta0);
  mpz_clear (beta1);
  mpz_clear (gamma0);
  mpz_clear (gamma1);
  mpz_clear (g0);
  mpz_clear (g1);

  l = realloc (l, *n * sizeof (unsigned long));

  return l;
}

static void
compsin (mpfi_t i)
{
  mpfi_const_pi (i);
  mpfi_mul_2exp (i, i, 1);
  mpfi_div_ui (i, i, 17);
  mpfi_sin (i, i);
}

int
main (int argc, char *argv[])
{
  unsigned long N = atoi (argv[1]), n;
  mp_prec_t p;
  mpfi_t i;
  mpz_t z1, z2;
  int st, st1;
  unsigned long *l;

  fprintf (stderr, "Using GMP %s and MPFR %s\n", gmp_version, mpfr_version);
  st = cputime ();

  /* initial estimate of the working precision */
  p = (mp_prec_t) (3.45 * (double) N + 17.0);
  fprintf (stderr, "Setting initial precision to %lu\n", p);

  mpfi_init (i);
  mpz_init (z1);
  mpz_init (z2);

  do
    {
      mpfi_set_prec (i, p);

      st1 = cputime ();
      compsin (i);
      fprintf (stderr, "Computing X took %dms\n", cputime () - st1);

      l = confrac (i, &n);
      fprintf (stderr, "found %lu partial quotients\n", n);

      if (n >= N)
	break;
      else
	{
          p += p / 10;
          fprintf (stderr, "Increasing precision to %lu\n", p);
	}
    }
  while (1);

  printf ("%lu\n", l[N - 1]);

  mpfi_clear (i);
  mpz_clear (z1);
  mpz_clear (z2);
  free (l);

  fprintf (stderr, "Cputime: %dms\n", cputime () - st);
  return 0;
}
