/* implement Lehmer's method (page 4, equation (3)) from
   http://web.comlab.ox.ac.uk/oucl/work/richard.brent/pub/pub166.html
*/

#include "many.h"

/* Error on alpha < 3.6 * 2^(-p) */
static unsigned long *
confrac (mpfr_t alpha, unsigned long *n, mp_prec_t prec)
{
  unsigned long *l, alloc, a;
  mpz_t beta0, beta1, gamma0, gamma1, g0, g1;
  mp_size_t n0;

  alloc = 1;
  l = malloc (alloc * sizeof (unsigned long));
  *n = 0; /* number of stored elements */

  mpz_init (beta0);
  mpz_init (beta1);
  mpz_init (gamma0);
  mpz_init (gamma1);
  mpz_init (g0);
  mpz_init (g1);

  mpfr_mul_2exp (alpha, alpha, prec, GMP_RNDN); /* err < 3.6 */
  mpfr_get_z (beta1, alpha, GMP_RNDD);
  mpz_sub_ui (gamma1, beta1, 3);
  mpz_add_ui (beta1, beta1, 4);
  mpz_set_ui (beta0, 1);
  mpz_mul_2exp (beta0, beta0, prec);
  mpz_set (gamma0, beta0);

  /* beta0/beta1 < 1/alpha < gamma0/gamma1 */

  while (1)
    {
      /* we must have gamma0 >= gamma1 */
      n0 = mpz_sizeinbase (gamma0, 2);
      n0 = (n0 >= mp_bits_per_limb) ? n0 - mp_bits_per_limb : 0;
      mpz_tdiv_q_2exp (g0, gamma0, n0);
      mpz_cdiv_q_2exp (g1, gamma1, n0);

      /* we have g0 = floor(gamma0/2^n0) and g1 = ceil(gamma1/2^n0)
         thus floor (g0/g1) <= floor(gamma0/gamma1) */

      a = mpz_get_ui (g0) / mpz_get_ui (g1);

      mpz_submul_ui (gamma0, gamma1, a);
      /* if a was too small, then gamma0 >= gamma1 */
      while (mpz_cmp (gamma0, gamma1) > 0)
        {
          mpz_sub (gamma0, gamma0, gamma1);
          a ++;
        }
      mpz_submul_ui (beta0, beta1, a);
      if (mpz_cmp_ui (beta0, 0) < 0 || mpz_cmp (beta0, beta1) >= 0)
        break;
      mpz_swap (gamma0, gamma1);
      mpz_swap (beta0, beta1);

      if (*n >= alloc)
        {
          alloc = 2 * alloc;
          l = realloc (l, alloc * sizeof (unsigned long));
        }
      l[*n] = a;
      (*n)++;

    }

  mpz_clear (beta0);
  mpz_clear (beta1);
  mpz_clear (gamma0);
  mpz_clear (gamma1);
  mpz_clear (g0);
  mpz_clear (g1);

  l = realloc (l, *n * sizeof (unsigned long));

  return l;
}

static void
compsin (mpfr_t s, mp_prec_t p)
{
  mpfr_t x, y, z;

  mpfr_set_prec (s, p);
  mpfr_inits2 (p, x, y, z, (void *) 0);

  /* Note: sqrt(x+eps) - sqrt(x) < eps / (2 * sqrt(x)) */

  mpfr_sqrt_ui (x, 17, GMP_RNDN); /* err < 2^(2-p) */
  mpfr_mul_2ui (z, x, 1, GMP_RNDN); /* err < 2^(3-p) */
  mpfr_ui_sub (y, 34, z, GMP_RNDN); /* err < (8+16) * 2^(-p) = 24 * 2^(-p) */
  mpfr_sqrt (z, y, GMP_RNDN); /* err < (2.4 + 4) * 2^(-p) = 6.4 * 2^(-p) */
  mpfr_mul_2ui (z, z, 1, GMP_RNDN); /* err < 12.8 * 2^(-p) */
  mpfr_ui_sub (x, 40, y, GMP_RNDN); /* err < (24+8) * 2^(-p) = 32 * 2^(-p) */
  mpfr_add (y, y, z, GMP_RNDN); /* err < (24+13+32) * 2^(-p) = 69 * 2^(-p) */
  mpfr_add_ui (z, z, 12, GMP_RNDN); /* err < (13+16) * 2^(-p) = 29 * 2^(-p) */
  mpfr_mul (z, z, x, GMP_RNDN); /* err < 1378 * 2^(-p) */
  mpfr_add_ui (z, z, 64, GMP_RNDN); /* err < 1634 * 2^(-p) */
  mpfr_mul_2ui (z, z, 1, GMP_RNDN); /* err < 3268 * 2^(-p) */
  mpfr_sqrt (s, z, GMP_RNDN); /* err < (60 + 16) * 2^(-p) = 76 * 2^(-p) */
  mpfr_sub (y, y, s, GMP_RNDN); /* err < (69+76+8) * 2^(-p) = 153 * 2^(-p) */
  mpfr_sqrt (s, y, GMP_RNDN); /* err < (26.5 + 2) * 2^(-p) = 28.5 * 2^(-p) */
  mpfr_div_2ui (s, s, 3, GMP_RNDN); /* err < 3.6 * 2^(-p) */

  mpfr_clears (x, y, z, (void *) 0);
}

int
main (int argc, char *argv[])
{
  unsigned long N = atoi (argv[1]), n;
  mp_prec_t p;
  mpfr_t s;
  mpz_t z1, z2;
  int st, st1;
  unsigned long *l;

  fprintf (stderr, "Using GMP %s and MPFR %s\n", gmp_version, mpfr_version);
  st = cputime ();

  /* initial estimate of the working precision */
  p = (mp_prec_t) (3.45 * (double) N + 17.0);
  fprintf (stderr, "Setting initial precision to %lu\n", p);

  mpfr_init (s);
  mpz_init (z1);
  mpz_init (z2);

  do
    {
      st1 = cputime ();
      compsin (s, p);
      fprintf (stderr, "Computing X took %dms\n", cputime () - st1);

      l = confrac (s, &n, p);
      fprintf (stderr, "found %lu partial quotients\n", n);

      if (n >= N)
	break;
      else
	{
          p += p / 10;
          fprintf (stderr, "Increasing precision to %lu\n", p);
	}
    }
  while (1);

  printf ("%lu\n", l[N - 1]);

  mpfr_clear (s);
  mpz_clear (z1);
  mpz_clear (z2);
  free (l);

  fprintf (stderr, "Cputime: %dms\n", cputime () - st);
  return 0;
}
