/* implement Lehmer's method (page 4, equation (3)) from
   http://web.comlab.ox.ac.uk/oucl/work/richard.brent/pub/pub166.html
   subquadratic implementation
*/

#include "many.h"

static void
compsin (mpfr_t s, mp_prec_t p)
{
  mpfr_t x, y, z;

  mpfr_set_prec (s, p);
  mpfr_inits2 (p, x, y, z, (void *) 0);

  /* Note: sqrt(x+eps) - sqrt(x) < eps / (2 * sqrt(x)) */

  mpfr_sqrt_ui (x, 17, GMP_RNDN); /* err < 2^(2-p) */
  mpfr_mul_2ui (z, x, 1, GMP_RNDN); /* err < 2^(3-p) */
  mpfr_ui_sub (y, 34, z, GMP_RNDN); /* err < (8+16) * 2^(-p) = 24 * 2^(-p) */
  mpfr_add_ui (z, z, 34, GMP_RNDN); /* err < (8+32) * 2^(-p) = 40 * 2^(-p) */
  mpfr_sqrt (y, y, GMP_RNDN); /* err < (2.4 + 4) * 2^(-p) = 6.4 * 2^(-p) */
  mpfr_sqrt (z, z, GMP_RNDN); /* err < (3.1 + 4) * 2^(-p) = 7.1 * 2^(-p) */
  mpfr_mul_2ui (z, z, 3, GMP_RNDN); /* err < 56.8 * 2^(-p) */
  mpfr_mul_ui (s, x, 6, GMP_RNDN); /* err < (24+16) * 2^(-p) = 40 * 2^(-p) */
  mpfr_add_ui (s, s, 34, GMP_RNDN); /* err < (40+32) * 2^(-p) = 72 * 2^(-p) */
  mpfr_sub (s, s, z, GMP_RNDN); /* err < (72 + 56.8) * 2^(-p) */
  mpfr_sub_ui (z, x, 1, GMP_RNDN); /* err < 2^(2-p) */
  mpfr_mul (z, z, y, GMP_RNDN); /* err < 48.3 * 2^(-p) */
  mpfr_add (s, s, z, GMP_RNDN); /* err < (128.8 + 48.3 + 16) * 2^(-p) */
  mpfr_mul_2ui (s, s, 1, GMP_RNDN); /* err < 387 * 2^(-p) */
  mpfr_sqrt (s, s, GMP_RNDN); /* err < (28.8 + 4) * 2^(-p) = 32.8 * 2^(-p) */
  mpfr_add (x, x, y, GMP_RNDN); /* err < 18.4 * 2^(-p) */
  mpfr_add (s, s, x, GMP_RNDN); /* err < 59.2 * 2^(-p) */
  mpfr_sub_ui (s, s, 1, GMP_RNDN); /* err < 59.2 * 2^(-p) */
  mpfr_div_2ui (s, s, 4, GMP_RNDN); /* err < 3.7 * 2^(-p) */
  mpfr_sqr (s, s, GMP_RNDN); /* err < 7.5 * 2^(-p) */
  mpfr_ui_sub (s, 1, s, GMP_RNDN); /* err < 7.5 * 2^(-p) */
  mpfr_sqrt (s, s, GMP_RNDN); /* err < 10.7 * 2^(-p) */

  mpfr_clears (x, y, z, (void *) 0);
}

/* compute lower/uppers approximations of sin(2*Pi/17), multiplied by 2^p */
static void
computeX (mpfr_t low, mpfr_t high)
{
  mp_prec_t p = mpfr_get_prec (low);
  compsin (low, p);
  /* the error is < 10.7 * 2^(-p) */
  mpfr_mul_2exp (low, low, p, GMP_RNDN); /* now error <= 10.7 */
  mpfr_add_ui (high, low, 11, GMP_RNDU);
  mpfr_sub_ui (low, low, 11, GMP_RNDD);
}

/* alpha is an interval enclosing the true value */
static unsigned long*
confrac_basecase (mpz_t beta0, mpz_t beta1, mpz_t gamma0, mpz_t gamma1,
                  unsigned long *n)
{
  unsigned long *l, alloc, a;
  mpz_t g0, g1;
  mp_size_t n0;

  alloc = 1;
  l = malloc (alloc * sizeof (unsigned long));
  *n = 0; /* number of stored elements */

  mpz_init (g0);
  mpz_init (g1);

  while (1)
    {
      /* we must have gamma0 >= gamma1 */
      n0 = mpz_sizeinbase (gamma0, 2);
      n0 = (n0 >= mp_bits_per_limb) ? n0 - mp_bits_per_limb : 0;
      mpz_tdiv_q_2exp (g0, gamma0, n0);
      mpz_cdiv_q_2exp (g1, gamma1, n0);

      if (mpz_cmp_ui (g1, 0) == 0)
        break;

      /* we have g0 = floor(gamma0/2^n0) and g1 = ceil(gamma1/2^n0)
         thus floor (g0/g1) <= floor(gamma0/gamma1) */

      a = mpz_get_ui (g0) / mpz_get_ui (g1);

      mpz_submul_ui (gamma0, gamma1, a);
      /* if a was too small, then gamma0 >= gamma1 */
      while (mpz_cmp (gamma0, gamma1) > 0)
        {
          mpz_sub (gamma0, gamma0, gamma1);
          a ++;
        }
      mpz_submul_ui (beta0, beta1, a);

      if (mpz_cmp_ui (beta0, 0) < 0 || mpz_cmp (beta0, beta1) >= 0)
        break;
      mpz_swap (gamma0, gamma1);
      mpz_swap (beta0, beta1);

      if (*n >= alloc)
        {
          alloc = 2 * alloc;
          l = realloc (l, alloc * sizeof (unsigned long));
        }
      l[*n] = a;
      (*n)++;

    }

  mpz_clear (g0);
  mpz_clear (g1);

  l = realloc (l, *n * sizeof (unsigned long));
  
  return l;
}

static void
matrix (mpz_t a, mpz_t b, mpz_t c, mpz_t d, unsigned long *l, unsigned long n)
{
  if (n == 1)
    {
      mpz_set_ui (a, 0);
      mpz_set_ui (b, 1);
      mpz_set_ui (c, 1);
      mpz_set_si (d, -l[0]);
    }
  else
    {
      unsigned long k = n / 2;
      mpz_t a0, b0, c0, d0, t;

      mpz_init (a0);
      mpz_init (b0);
      mpz_init (c0);
      mpz_init (d0);
      mpz_init (t);

      matrix (a0, b0, c0, d0, l, k);
      matrix (a, b, c, d, l + k, n - k);
      
      /* now compute [a,b;c,d] * [a0,b0;c0,d0] = 
         [c0*b + a0*a, d0*b + b0*a; a0*c + c0*d, b0*c + d0*d] */

      mpz_swap (t, a);       /* save a */
      mpz_mul (a, a0, t);    /* a0*a */
      mpz_addmul (a, c0, b); /* c0*b + a0*a */
      mpz_mul (b, d0, b);    /* d0*b */
      mpz_addmul (b, b0, t); /* d0*b + b0*a */
      mpz_swap (t, c);       /* save c */
      mpz_mul (c, a0, t);    /* a0*c */
      mpz_addmul (c, c0, d); /* a0*c + c0*d */
      mpz_mul (d, d0, d);    /* d0*d */
      mpz_addmul (d, b0, t); /* b0*c + d0*d */
      
      mpz_clear (a0);
      mpz_clear (b0);
      mpz_clear (c0);
      mpz_clear (d0);
      mpz_clear (t);
    }
}

/* [u;v] <- [a,b;c,d] * [u;v] */
static void
mul (mpz_t a, mpz_t b, mpz_t c, mpz_t d, mpz_t u, mpz_t v)
{
  mpz_t t;
  mpz_init (t);
  mpz_swap (t, u);
  mpz_mul (u, a, t);
  mpz_addmul (u, b, v);
  mpz_mul (v, d, v);
  mpz_addmul (v, c, t);
  mpz_clear (t);
}

#define THRESHOLD 7000 /* 5000: 32405ms, 6000: 32271ms, 7000:31951ms */

/* alpha is an interval enclosing the true value */
static unsigned long*
confrac_fast (mpz_t beta0, mpz_t beta1, mpz_t gamma0, mpz_t gamma1,
              unsigned long *n)
{
  mp_size_t l = mpz_sizeinbase (gamma0, 2), k;
  mpz_t b0, b1, g0, g1;
  unsigned long *l1, *l2, n2, i;

  if (mpz_cmp (gamma0, beta0) < 0) abort();

  if (l < THRESHOLD)
    return confrac_basecase (beta0, beta1, gamma0, gamma1, n);

  k = 3 * l / 4;

  mpz_init (b0);
  mpz_init (b1);
  mpz_init (g0);
  mpz_init (g1);

  mpz_tdiv_q_2exp (b0, beta0, k);
  mpz_cdiv_q_2exp (b1, beta1, k);
  mpz_cdiv_q_2exp (g0, gamma0, k);
  mpz_tdiv_q_2exp (g1, gamma1, k);

  l1 = confrac_fast (b0, b1, g0, g1, n);
  if (*n == 0)
    {
      l1 = confrac_basecase (beta0, beta1, gamma0, gamma1, n);
      goto end;
    }

  matrix (b0, b1, g0, g1, l1, *n);
  
  mul (b0, b1, g0, g1, beta0, beta1);
  mul (b0, b1, g0, g1, gamma0, gamma1);

  /* swap already done between beta0 and beta1, so beta1 is the last term */
  if (mpz_cmp_ui (beta1, 0) < 0 || mpz_cmp (beta1, beta0) >= 0)
    goto end;

  if (mpz_cmp (gamma0, beta0) < 0)
    {
      mpz_swap (beta0, gamma0);
      mpz_swap (beta1, gamma1);
    }

  l2 = confrac_fast (beta0, beta1, gamma0, gamma1, &n2);

  l1 = realloc (l1, (*n + n2) * sizeof (unsigned long));
  for (i = 0; i < n2; i++)
    l1[*n + i] = l2[i];
  *n += n2;
  free (l2);

 end:
  mpz_clear (b0);
  mpz_clear (b1);
  mpz_clear (g0);
  mpz_clear (g1);

  return l1;
}

int
main (int argc, char *argv[])
{
  unsigned long N = atoi (argv[1]), n;
  mp_prec_t p;
  mpfr_t low, high;
  int st, st1;
  unsigned long *l;
  mpz_t beta0, beta1, gamma0, gamma1;

  mpz_init (beta0);
  mpz_init (beta1);
  mpz_init (gamma0);
  mpz_init (gamma1);

  fprintf (stderr, "Using GMP %s and MPFR %s\n", gmp_version, mpfr_version);
  st = cputime ();

  /* initial estimate of the working precision */
  p = (mp_prec_t) (argc > 2) ? atoi (argv[2]) : (3.43 * (double) N + 10.0);
  fprintf (stderr, "Setting initial precision to %lu\n", p);
  
  mpfr_init (low);
  mpfr_init (high);

  do
    {
      mpfr_set_prec (low, p);
      mpfr_set_prec (high, p);

      st1 = cputime ();
      computeX (low, high);
      fprintf (stderr, "Computing X took %dms\n", cputime () - st1);

      st1 = cputime ();
      mpfr_get_z (beta1, high, GMP_RNDU);
      mpfr_get_z (gamma1, low, GMP_RNDD);
      mpz_set_ui (beta0, 1);
      mpz_mul_2exp (beta0, beta0, p);
      mpz_set (gamma0, beta0);

      /* beta0/beta1 < 1/alpha < gamma0/gamma1 */

      l = confrac_fast (beta0, beta1, gamma0, gamma1, &n);
      fprintf (stderr, "Continued fraction took %dms\n", cputime () - st1);
      fprintf (stderr, "found %lu partial quotients\n", n);

      if (n >= N)
	break;
      else
	{
          p += p / 10;
          fprintf (stderr, "Increasing precision to %lu\n", p);
	}
    }
  while (1);

  printf ("%lu\n", l[N - 1]);

  mpfr_clear (low);
  mpfr_clear (high);
  free (l);
  mpz_clear (beta0);
  mpz_clear (beta1);
  mpz_clear (gamma0);
  mpz_clear (gamma1);

  fprintf (stderr, "Cputime: %dms\n", cputime () - st);
  return 0;
}
