/* check.c -- check the library via sample integrals.

Copyright 2005 Laurent Fousse <laurent@komite.net>

This file is part of the CRQ Library.

The CRQ Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or (at your
option) any later version.

The CRQ Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with the CRQ Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Place, Fifth Floor, Boston,
MA 02110-1301, USA. */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "crq.h"
#include "crq-impl.h"
#include "utils.h"

#define WANT_TIMINGS

extern mpfr_t max_dyi;

enum { GL, NC };

void sin_wrapper (mpfr_t res, mpfr_t x, mp_prec_t rndm)
{
    mpfr_sin (res, x, rndm);
}

void square_wrapper (mpfr_t res, mpfr_t x, mp_prec_t rndm)
{
    mpfr_mul (res, x, x, rndm);
}

void sinsin (mpfr_t res, mpfr_t x, mp_prec_t rndm)
{
    mpfr_sin (res, x, rndm);
    mpfr_sin (res, res, rndm);
}

void one_wrapper (mpfr_t res, mpfr_t x, mp_prec_t rndm)
{
    mpfr_set_ui (res, 1, GMP_RNDN);
}

void exp_wrapper (mpfr_t res, mpfr_t x, mp_prec_t rndm)
{
#if 1
    printf ("On va calculer exp avec x = ");
    mpfr_out_str (stdout, DBASE, 0, x, GMP_RNDN);
    printf ("\n");
#endif
    mpfr_exp (res, x, rndm);
    if (mpfr_nan_p (res))
	printf ("Quelque chose de pourri au royaume du Danemark.\n");
}

void nth_deriv_bound_exp (mpfr_t res, mpfr_t a, mpfr_t b, unsigned int n)
{
    mpfr_exp (res, b, GMP_RNDU);
}

/* t*log(1+t) */
void bailey1 (mpfr_t res, mpfr_t t, mp_prec_t rndm)
{
    mpfr_log1p (res, t, rndm);
    mpfr_mul (res, res, t, rndm);
}

/* sqrt(1-t^2) */
void bailey2 (mpfr_t res, mpfr_t t, mp_prec_t rndm)
{
    mpfr_sqr (res, t, rndm);
    mpfr_ui_sub (res, 1, res, rndm);
    mpfr_sqrt (res, res, rndm);
}

int main (int argc, char *argv[])
{
    unsigned int n, m;
    mp_prec_t prec, wp = 0;

    mpfr_t a, b;
    mpfr_t res;
    mpfr_t exact_value;
    mpfr_t total_error, math_error, roundoff_error, eval_error;
    mpfr_t nan;
    mpfr_t foo, bar;
    mpfr_t trois;
#ifdef WANT_TIMINGS
    int total_time = 0;
#endif
    int options = 0;
    int testcase = 0;
    int method = GL;
    struct crq_stats stats;

    stats.usp_time = 0;
    stats.refine_time = 0;
    stats.coeff_time = 0;
    stats.newton_good = 0;
    stats.newton_bad = 0;
    stats.dich_prep_time = 0;
    stats.dicho_time = 0;
    stats.newton_failures = 0;
 

    m = 1;
    n = 42;
    mpfr_init (nan);
    mpfr_set_nan (nan);

    while (argc > 1 && argv[1][0] == '-') {
	if (argc >= 3 && strcmp (argv[1], "-p") == 0) {
	    prec = atoi (argv[2]);	/* target precision */
	    mpfr_set_default_prec (prec);
	    argv += 2;
	    argc -= 2;
	}
	else if (argc >= 3 && strcmp (argv[1], "-wp") == 0) {
	    wp = atoi (argv[2]);	/* working precision */
	    argv += 2;
	    argc -= 2;
	}
	else if (argc >= 3 && strcmp (argv[1], "-test") == 0) {
	    testcase = atoi (argv[2]);	/* which test */
	    argv += 2;
	    argc -= 2;
	}
	else if (argc >= 3 && strcmp (argv[1], "-n") == 0) {
	    n = atoi (argv[2]);	/* Method with n points */
	    argv += 2;
	    argc -= 2;
	}
	else if (argc >= 3 && strcmp (argv[1], "-m") == 0) {
	    m = atoi (argv[2]);	/* number of intervals */
	    argv += 2;
	    argc -= 2;
	}
	else if (argc >= 2 && strcmp (argv[1], "-newton") == 0) {
	    if (options & 1)
		options ^= 1;
	    argv += 1;
	    argc -= 1;
	}
	else if (argc >= 2 && strcmp (argv[1], "-nc") == 0) {
	    method = NC;
	    argv += 1;
	    argc -= 1;
	}
	else if (argc >= 2 && strcmp (argv[1], "-gl") == 0) {
	    method = GL;
	    argv += 1;
	    argc -= 1;
	}
	else if (argc >= 2 && strcmp (argv[1], "-newton_int") == 0) {
	    options |= 4;
	    argv += 1;
	    argc -= 1;
	}
	else if (argc >= 2 && strcmp (argv[1], "-dicho") == 0) {
	    options |= 1;
	    argv += 1;
	    argc -= 1;
	}
	else if (argc >= 2 && strcmp (argv[1], "-dicho_trunc") == 0) {
	    options |= 8;
	    argv += 1;
	    argc -= 1;
	}
	else if (argc >= 2 && strcmp (argv[1], "-usp2") == 0) {
	    options |= 2;
	    argv += 1;
	    argc -= 1;
	}
	else {
	    fprintf (stderr, "Invalid option: %s\n", argv[1]);
	    exit (1);
	}
    }

    if (wp == 0)
	wp = mpfr_get_default_prec ();

    mpfr_init2 (a, wp);
    mpfr_init2 (b, wp);
    mpfr_init2 (res, wp);
    mpfr_init2 (exact_value, wp);

    mpfr_set_ui (a, 0, GMP_RNDN);
    mpfr_set_ui (b, 3, GMP_RNDN);

    printf ("%d [n]\n%d [m]\n%lu [wp]\n", n, m, wp);
    if (DICHO(options))
	printf ("[dicho]\n");
    if (USP_COUPLE(options))
	printf ("[usp_couple]\n");
    if (NEWTON_INT(options))
	printf ("[newton_int]\n");
    if (NEWTON(options))
	printf ("[newton]\n");

    switch (testcase) {
	case 0: /* f = 1, [a, b] = [0, 3] */
	    printf ("f(x) = 1, [a, b] = [0, 3]\n");
#ifdef WANT_TIMINGS
	    total_time = cputime ();
#endif
	    if (method == GL)
		compose_gl_stats (res, nan, a, b, m, n, one_wrapper, nan, wp, options, NULL, &stats);
	    else
		compose_nc (res, a, b, m, n, one_wrapper, nan, nan, wp);
#ifdef WANT_TIMINGS
	    total_time = cputime () - total_time;
#endif
	    printf ("val = ");
	    mpfr_out_str (NULL, DBASE, 0, res, GMP_RNDN); printf ("\n");
	    mpfr_set_ui (exact_value, 3, GMP_RNDN);
	    printf ("Exact val: ");
	    mpfr_out_str (NULL, DBASE, 0, exact_value, GMP_RNDN); printf ("\n");
	    mpfr_sub (exact_value, exact_value, res, GMP_RNDN);
	    mpfr_abs (exact_value, exact_value, GMP_RNDN);
	    mpfr_out_str (NULL, DBASE, 0, exact_value, GMP_RNDN);
	    printf (" [measured error]\n");
	    break;

	case 1: /* f = exp, [a, b] = [0, 3] */
	    printf ("f(x) = exp(x), [a, b] = [0, 3]\n");
#ifdef WANT_TIMINGS
	    total_time = cputime ();
#endif
	    mpfr_init (trois);
	    mpfr_init (total_error);
	    mpfr_init (roundoff_error);
	    mpfr_init (math_error);
	    mpfr_init (eval_error);
	    mpfr_set_ui (trois, 3, GMP_RNDN);
	    if (method == GL)
		compose_gl_stats (res, eval_error, a, b, m, n, exp_wrapper, trois, wp, options, NULL, &stats);
	    else
		compose_nc (res, a, b, m, n, exp_wrapper, nan, nan, wp);
#ifdef WANT_TIMINGS
	    total_time = cputime () - total_time;
#endif
	    printf ("val = ");
	    mpfr_out_str (NULL, DBASE, 0, res, GMP_RNDN); printf ("\n");
	    mpfr_set_prec (exact_value, wp + 50);
	    mpfr_set_ui (exact_value, 3, GMP_RNDN);
	    mpfr_exp (exact_value, exact_value, GMP_RNDN);
	    mpfr_sub_ui (exact_value, exact_value, 1, GMP_RNDN);
	    printf ("Exact val: ");
	    mpfr_out_str (NULL, DBASE, 0, exact_value, GMP_RNDN); printf ("\n");
	    mpfr_sub (exact_value, exact_value, res, GMP_RNDN);
	    mpfr_abs (exact_value, exact_value, GMP_RNDN);
	    mpfr_out_str (NULL, DBASE, 0, exact_value, GMP_RNDN);
	    printf (" [measured error]\n");
	    mpfr_set_ui (exact_value, 3, GMP_RNDN);
	    mpfr_exp (exact_value, exact_value, GMP_RNDN);
	    mpfr_init (foo);
	    mpfr_init (bar);
	    gl_error (math_error, a, b, 1, n, exact_value);
	    mpfr_out_str (NULL, DBASE, 0, math_error, GMP_RNDN);
	    printf (" [method error]\n");
	    mpfr_set (total_error, math_error, GMP_RNDU);
	    /* Static error */
	    mpfr_set_ui (foo, 9, GMP_RNDN);
	    mpfr_mul_2si (foo, foo, -1, GMP_RNDN);
	    mpfr_set_ui (bar, 3, GMP_RNDN);
	    mpfr_mul_2si (bar, bar, -wp, GMP_RNDN);
	    mpfr_add (foo, foo, bar, GMP_RNDU);
	    mpfr_setulp (bar, res);
	    mpfr_mul (foo, foo, bar, GMP_RNDU);
	    mpfr_out_str (NULL, DBASE, 0, foo, GMP_RNDN);
	    printf (" [static error]\n");
	    mpfr_add (total_error, total_error, foo, GMP_RNDU);
	    mpfr_set (roundoff_error, foo, GMP_RNDU);

	    mpfr_out_str (NULL, DBASE, 0, eval_error, GMP_RNDN);
	    printf (" [eval error]\n");
	    mpfr_add (total_error, total_error, eval_error, GMP_RNDU);
	    mpfr_add (roundoff_error, roundoff_error, foo, GMP_RNDU);
	    mpfr_out_str (NULL, DBASE, 0, total_error, GMP_RNDN);
	    printf (" [total error]\n");
	    mpfr_out_str (NULL, DBASE, 0, roundoff_error, GMP_RNDN);
	    printf (" [roundoff error]\n");
	    printf ("%ld [correct bits]\n", mpfr_get_exp (res) -
					   mpfr_get_exp (total_error));

	    if (mpfr_cmp (math_error, roundoff_error) > 0)
		printf ("Insufficient order\n");
	    else
		printf ("Insufficient prec\n");

	    mpfr_clear (foo);
	    break;

	case 2: /* f(t) = t*log(1+t), [a, b] = [0, 1] */
	    printf ("f(x) = x*log(1+x), [a, b] = [0, 1]\n");
#ifdef WANT_TIMINGS
	    total_time = cputime ();
#endif
	    mpfr_set_ui (a, 0, GMP_RNDN);
	    mpfr_set_ui (b, 1, GMP_RNDN);
	    if (method == GL)
		compose_gl_stats (res, nan, a, b, m, n, bailey1, nan, wp, options, NULL, &stats);
	    else
		compose_nc (res, a, b, m, n, bailey1, nan, nan, wp);
#ifdef WANT_TIMINGS
	    total_time = cputime () - total_time;
#endif
	    printf ("val = ");
	    mpfr_out_str (NULL, DBASE, 0, res, GMP_RNDN); printf ("\n");
	    mpfr_set_ui (exact_value, 1, GMP_RNDN);
	    mpfr_mul_2si (exact_value, exact_value, -2, GMP_RNDN);
	    printf ("Exact val: ");
	    mpfr_out_str (NULL, DBASE, 0, exact_value, GMP_RNDN); printf ("\n");
	    mpfr_sub (exact_value, exact_value, res, GMP_RNDN);
	    mpfr_abs (exact_value, exact_value, GMP_RNDN);
	    mpfr_out_str (NULL, DBASE, 0, exact_value, GMP_RNDN);
	    printf (" [measured error]\n");
	    break;

	case 3: /* f(t) = sqrt(1-t^2), [a, b] = [0, 1] */
	    printf ("f(t) = sqrt(1-t^2), [a, b] = [0, 1]\n");
#ifdef WANT_TIMINGS
	    total_time = cputime ();
#endif
	    mpfr_set_ui (a, 0, GMP_RNDN);
	    mpfr_set_ui (b, 1, GMP_RNDN);
	    if (method == GL)
		compose_gl_stats (res, nan, a, b, m, n, bailey2, nan, wp, options, NULL, &stats);
	    else
		compose_nc (res, a, b, m, n, bailey2, nan, nan, wp);
#ifdef WANT_TIMINGS
	    total_time = cputime () - total_time;
#endif
	    printf ("val = ");
	    mpfr_out_str (NULL, DBASE, 0, res, GMP_RNDN); printf ("\n");
	    mpfr_const_pi (exact_value, GMP_RNDN);
	    mpfr_mul_2si (exact_value, exact_value, -2, GMP_RNDN);
	    printf ("Exact val: ");
	    mpfr_out_str (NULL, DBASE, 0, exact_value, GMP_RNDN); printf ("\n");
	    mpfr_sub (exact_value, exact_value, res, GMP_RNDN);
	    mpfr_abs (exact_value, exact_value, GMP_RNDN);
	    mpfr_out_str (NULL, DBASE, 0, exact_value, GMP_RNDN);
	    printf (" [measured error]\n");
	    break;

	case 4: /* f(t) = sin(sin(x)), [a, b] = [0, 1] */
	    printf ("f(t) = sin(sin(x)), [a, b] = [0, 1]\n");
#ifdef WANT_TIMINGS
	    total_time = cputime ();
#endif
	    mpfr_set_ui (a, 0, GMP_RNDN);
	    mpfr_set_ui (b, 1, GMP_RNDN);
	    if (method == GL)
		compose_gl_stats (res, nan, a, b, m, n, sinsin, nan, wp, options, NULL, &stats);
	    else
		compose_nc (res, a, b, m, n, sinsin, nan, nan, wp);
#ifdef WANT_TIMINGS
	    total_time = cputime () - total_time;
#endif
	    printf ("val = ");
	    mpfr_out_str (NULL, DBASE, 0, res, GMP_RNDN); printf ("\n");
	    break;
    }

#ifdef WANT_TIMINGS
    printf ("%d [Total CPU time]\n", total_time);
#endif
    printf ("%d [Uspensky time]\n", stats.usp_time);
    printf ("%d [Root refinement time]\n", stats.refine_time);
    printf ("%d [Coefficient computation time]\n", stats.coeff_time);
    printf ("%d [Useful newton time]\n", stats.newton_good);
    printf ("%d [Useless newton time]\n", stats.newton_bad);
    printf ("%d [dicho preparation time]\n", stats.dich_prep_time);
    printf ("%d [dicho time]\n", stats.dicho_time);
    printf ("%d [newton failures]\n", stats.newton_failures);

    mpfr_clear (a);
    mpfr_clear (b);
    mpfr_clear (exact_value);
    mpfr_clear (res);
    return 0;
}
