/* nc_gen.c -- generic method for Newton-Cotes.

Copyright 2005 Laurent Fousse <laurent@komite.net>

This file is part of the CRQ Library.

The CRQ Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or (at your
option) any later version.

The CRQ Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with the CRQ Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Place, Fifth Floor, Boston,
MA 02110-1301, USA. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gmp.h>
#include "crq.h"
#include "crq-impl.h"
#include <assert.h>
#include <mpfr.h>

#include "utils.h"

#define FUNCTION exp_wrapper
#define LEFT_BOUND 0
#define RIGHT_BOUND 3
/* #define DEBUG_EXACT */
/* #define TEST2 */
/*#define DTIS*/

#ifdef DEBUG_EXACT

void mpfr_add_exact (mpfr_t ret, mpfr_t a, mpfr_t b)
{
    mp_prec_t needed;

    if (mpfr_nan_p (a) || mpfr_nan_p (b) || mpfr_nan_p (ret)) {
	printf ("Duh. NAN.\n");
	abort ();
    }
    if mpfr_zero_p (a) {
	mpfr_set_prec (ret, mpfr_get_prec (b));
	mpfr_set (ret, b, GMP_RNDN);
	return;
    }
    if mpfr_zero_p (b) {
	mpfr_set_prec (ret, mpfr_get_prec (a));
	mpfr_set (ret, a, GMP_RNDN);
	return;
    }

    /* Grossly overestimate needed precision */
    if (mpfr_get_exp (a) > mpfr_get_exp (b)) {
	needed = mpfr_get_exp (a) - mpfr_get_exp (b);
	if (mpfr_get_prec (a) > mpfr_get_prec (b))
	    needed += mpfr_get_prec (a);
	else
	    needed += mpfr_get_prec (b);
    }
    else {
	needed = mpfr_get_exp (b) - mpfr_get_exp (a);
	if (mpfr_get_prec (a) > mpfr_get_prec (b))
	    needed += mpfr_get_prec (a);
	else
	    needed += mpfr_get_prec (b);
    }
    needed++;

    mpfr_prec_round (ret, needed, GMP_RNDN);
    mpfr_add (ret, a, b, GMP_RNDN);
    if (mpfr_nan_p (ret)) {
	printf ("There is a pestilence upon this land. Nothing is sacred\n");
	abort ();
    }
}

#endif

/* unused */
#if 0
void mpfr_sum_no_loop (mpfr_t res, mpfr_ptr tab, unsigned int n) 
{
    mpfr_srcptr *perm;

    perm = malloc (n * sizeof (*perm));
    mpfr_count_sort (tab, n, perm);

    free (perm);
}
#endif


/* assume b[0]...b[n/2] and d are initialized
   Return b[] and d such that b[k]/d is the k-th Newton-Cotes coefficient
   of order n, i.e. (-1)^(n-k)/k!/(n-k)!*int(t*(t-1)*...*(t-n)/(t-k),t=0..n).
   
   Remark: we don't need b[n/2+1]...b[n] since b[k]=b[n-k].
*/
void compute_bnk (mpz_t * b, mpz_t d, unsigned long n)
{
    unsigned long k, g, j;
    mpz_t num, den, pow_n, *p, fac_n;

    mpz_init (num);
    mpz_init (den);
    mpz_init (pow_n);
    mpz_init (fac_n);

    /* compute (multiple of) common denominator */
    mpz_fac_ui (fac_n, n);
    mpz_set_ui (num, 1);
    for (k = 2; k <= n; k++)
	mpz_lcm_ui (num, num, k);
    mpz_divexact (num, fac_n, num);	/* n! / lcm(1, 2, ..., n) */
    mpz_mul_ui (num, num, n);
    mpz_set_ui (pow_n, n);
    mpz_set_ui (den, 1);
    for (k = 1; k <= n; k++) {
	mpz_mul_ui (pow_n, pow_n, n);	/* pow_n = n^(k+1) */
	g = mpz_gcd_ui (NULL, pow_n, k + 1);
	mpz_divexact_ui (b[0], pow_n, g);
	mpz_gcd (num, num, b[0]);
	mpz_lcm_ui (den, den, (k + 1) / g);
    }
    mpz_gcd (num, num, den);
    mpz_divexact (den, den, num);

    p = (mpz_t *) malloc ((n + 1) * sizeof (mpz_t));
    for (k = 0; k <= n; k++)
	mpz_init (p[k]);
    /* put coefficients of (t-1)*(t-2)*...*(t-n) in p */
    mpz_set_ui (p[0], 1);
    for (k = 1; k <= n; k++) {
	/* p[0..k-1] contains the coefficients of (t-1)*...*(t-(k-1)) */
	mpz_set_ui (p[k], 0);
	for (j = k; j > 0; j--) {
	    mpz_mul_ui (p[j], p[j], k);
	    mpz_sub (p[j], p[j - 1], p[j]);
	}
	mpz_mul_si (p[0], p[0], -k);
    }
    /* compute b[0] */
    mpz_set (pow_n, den);
    mpz_set_ui (b[0], 0);
    for (j = 0; j <= n; j++) {
	mpz_mul_ui (pow_n, pow_n, n);	/* pow_n = den * n^(j+1) */
	mpz_mul (num, p[j], pow_n);
	mpz_divexact_ui (num, num, j + 1);
	mpz_add (b[0], b[0], num);
    }
    if (n % 2)
	mpz_neg (b[0], b[0]);

    /* compute b[1]...b[n/2] */
    mpz_set_ui (d, 1);		/* d = n!/k!/(n-k)! */
    for (k = 1; k <= n / 2; k++) {
	mpz_mul_ui (d, d, n - (k - 1));
	mpz_divexact_ui (d, d, k);
	/* multiply by t-k */
	for (j = n - 1; j > 0; j--)
	    mpz_addmul_ui (p[j], p[j + 1], k);
	/* divide by t-(k-1) */
	mpz_set_ui (p[0], 0);
	for (j = 0; j < n; j++)
	    mpz_submul_ui (p[j], p[j + 1], k - 1);
	/* set b[k] */
	mpz_set (pow_n, den);
	mpz_set_ui (b[k], 0);
	for (j = 0; j <= n; j++) {
	    mpz_mul_ui (pow_n, pow_n, n);	/* pow_n = den * n^(j+1) */
	    mpz_mul (num, p[j], pow_n);
	    mpz_divexact_ui (num, num, j + 1);
	    mpz_add (b[k], b[k], num);
	}
	if ((n - k) % 2)
	    mpz_neg (b[k], b[k]);
	mpz_mul (b[k], b[k], d);
    }

    for (k = 0; k <= n; k++)
	mpz_clear (p[k]);
    free (p);

    mpz_mul (d, fac_n, den);

    /* compute smallest denominator */
    mpz_gcd (den, d, b[0]);
    for (k = 1; k <= n / 2; k++)
	mpz_gcd (den, den, b[k]);
    mpz_divexact (d, d, den);
    for (k = 0; k <= n / 2; k++)
	mpz_divexact (b[k], b[k], den);

    mpz_clear (num);
    mpz_clear (den);
    mpz_clear (pow_n);
    mpz_clear (fac_n);
}

/* Error bound for NC generic rule for n points on [a, b].
 * This accounts only for the method error.
 * M is a bound of |f^(n)| for even n and of |f^(n+1)| for odd n
 * on [a, b].
 */

void nc_closed_error (mpfr_t res, mpfr_t a, mpfr_t b, unsigned int n, mpfr_t M)
{
    mpfr_t h;
    unsigned int power;

    mpfr_init (h);

    mpfr_sub (h, b, a, GMP_RNDU);
    mpfr_div_ui (h, h, n - 1, GMP_RNDU);

    if (n & 1)
	power = n + 2;
    else
	power = n + 1;

    mpfr_pow_ui (res, h, power, GMP_RNDU);
    mpfr_mul (res, res, M, GMP_RNDU);
    /* with odd n, |c_n| <= 1/8
     * with even n, |c_n| <= 1/4
     */

    if (n & 1)
	mpfr_mul_2si (res, res, -3, GMP_RNDU);
    else
	mpfr_mul_2si (res, res, -2, GMP_RNDU);
    mpfr_clear (h);
}
    
/* Closed NC generic rule for n points on [a, b] 
   wp is the working precision
 */
void
gen_nc_closed (mpfr_t errb, mpfr_t res, mpfr_t a, mpfr_t b, unsigned int n,
	       mpz_t * coeffs, mpz_t d, aqfunc_t f, mpfr_t m, mp_prec_t wp,
	       mpfr_t *fxs[])
{
    mpfr_t D;
    mpfr_t fx;
    mpfr_t x, foo, bar, apb;
    mpfr_t acc;
    mpfr_t t1, t2, err_stat, err_diff, err_eval;
    mpz_t gee;
    unsigned int i;
    mpfr_t curr_deltati, max_deltati;
#ifdef DTIS
    mpfr_t dtis;
#endif

    assert (n >= 2);

    mpfr_init2 (acc, wp);
    mpfr_set_ui (acc, 0, GMP_RNDN);

    mpfr_init2 (x, wp);
    mpfr_init2 (foo, wp);
    mpfr_init2 (bar, wp);
    mpfr_init2 (curr_deltati, wp);
    mpfr_init2 (max_deltati, wp);
#ifdef DTIS
    mpfr_init2 (dtis, wp);
#endif
    mpfr_set (x, a, GMP_RNDN);

    mpfr_init2 (fx, wp);
    mpfr_init2 (apb, wp);
    mpfr_add (apb, a, b, GMP_RNDN);

    for (i = 0; i < n/2; i++) {
	f (*fxs[i], x, mpfr_get_prec (x));
	/* Compute x' such that x + x' = a + b */
	mpfr_sub (x, apb, x, GMP_RNDN);
	f (foo, x, mpfr_get_prec (x));
	mpfr_add (*fxs[i], *fxs[i], foo, GMP_RNDN);
	mpfr_mul_z (*fxs[i], *fxs[i], coeffs[i], GMP_RNDN);
	/* We need to keep track of the max error bound on
	 * ti = f(xi) * ni. It's not merely given by ulp(ti) */
	mpfr_setulp (curr_deltati, *fxs[i]);
	mpfr_mul_ui (curr_deltati, curr_deltati, 3, GMP_RNDN);
	mpfr_mul_2si (curr_deltati, curr_deltati, -1, GMP_RNDN);
	/* 3/2 ulp(ti) */

	/* Duh. ulp(0) fails of course */
	if (mpfr_zero_p (x))
	    mpfr_set_ui (foo, 0, GMP_RNDN);
	else
	    mpfr_setulp (foo, x);
	mpfr_mul_ui (foo, foo, 6, GMP_RNDU);
	if (mpz_sgn (coeffs[i]) > 0) /* abs */
	    mpfr_mul_z (foo, foo, coeffs[i], GMP_RNDU);
	else {
	    mpfr_mul_z (foo, foo, coeffs[i], GMP_RNDD);
	    mpfr_abs (foo, foo, GMP_RNDU);
	}
	
	mpfr_mul (foo, foo, m, GMP_RNDU);
	/* foo = 6 ni m ulp(xi) */
	
	mpfr_add (curr_deltati, curr_deltati, foo, GMP_RNDU);

#ifdef DTIS
	if (i == 0)
	    mpfr_set (dtis, curr_deltati, GMP_RNDN);
	else
	    mpfr_add (dtis, dtis, curr_deltati, GMP_RNDU);
#else
	if (i == 0)
	    mpfr_set (max_deltati, curr_deltati, GMP_RNDN);
	if (mpfr_cmp (curr_deltati, max_deltati) > 0) {
	    mpfr_set (max_deltati, curr_deltati, GMP_RNDN);
	}
#endif
#ifdef DEBUG
	mpfr_out_str (stdout, DBASE, 5, curr_deltati, GMP_RNDN);
	printf (" [curdeltati]\n");
	mpfr_out_str (stdout, DBASE, 5, *fxs[i], GMP_RNDN);
	printf (" [curti]\n");
    #ifdef DTIS
	mpfr_out_str (stdout, DBASE, 5, dtis, GMP_RNDN);
	printf (" [curdtis]\n");
    #endif
#endif
	/* Compute new x */
	mpfr_mul_ui (foo, a, n - i - 2, GMP_RNDN);
	mpfr_mul_ui (bar, b, i + 1, GMP_RNDN);
	mpfr_add (x, foo, bar, GMP_RNDN);
	mpfr_div_ui (x, x, n - 1, GMP_RNDN);
    }
    if (n & 1) {
	mpfr_mul_2si (x, apb, -1, GMP_RNDN);
	f (*fxs[n/2], x, mpfr_get_prec (x));
	mpfr_mul_z (*fxs[n/2], *fxs[n/2], coeffs[n/2], GMP_RNDN);
	/* TODO : deltati... */
    }

    /* TODO : directly calling mpfr_sum is not suitable here */
    mpfr_sum (acc, (mpfr_ptr *) fxs, (n+1)/2, GMP_RNDN);

    mpfr_init2 (D, wp);

    /* TODO: Look for lost significant bits here. */
    mpfr_sub (D, b, a, GMP_RNDN);

    /* Error bound */

    mpfr_init (t1);
    mpfr_init (t2);
    mpfr_init (err_eval);
    mpfr_init (err_stat);
    mpfr_init (err_diff);

    mpfr_set_ui (t1, 43, GMP_RNDU);
    mpfr_mul_2si (t1, t1, -1, GMP_RNDU);

    mpfr_set_ui (t2, 21, GMP_RNDU);
    mpfr_mul_2si (t2, t2, -wp, GMP_RNDU);

    mpfr_add (err_stat, t1, t2, GMP_RNDU);


#ifdef DTIS
    mpfr_set_ui (err_eval, 1, GMP_RNDU);
#else
    mpfr_set_ui (err_eval, n, GMP_RNDU);
#endif
    mpfr_div_ui (err_eval, err_eval, n - 1, GMP_RNDU);
    mpfr_mul_ui (err_eval, err_eval, 6, GMP_RNDU);
    mpfr_div_z (err_eval, err_eval, d, GMP_RNDU);
    mpfr_mul (err_eval, err_eval, D, GMP_RNDU);
    mpfr_set_ui (t1, 1, GMP_RNDU);
    mpfr_mul_2si (t1, t1, -wp, GMP_RNDU);
    mpfr_add_ui (t1, t1, 1, GMP_RNDU);
    mpfr_mul (err_eval, err_eval, t1, GMP_RNDU);
#ifdef DTIS
    mpfr_mul (err_eval, dtis, err_eval, GMP_RNDU);
#else
    mpfr_mul (err_eval, max_deltati, err_eval, GMP_RNDU);
#endif

    mpz_init (gee);
    mpz_mul_ui (gee, d, n - 1);
    mpfr_div_z (res, acc, gee, GMP_RNDN);
    mpz_clear (gee);

    mpfr_setulp (t1, b);
    mpfr_setulp (t2, a);
    mpfr_add (t1, t1, t2, GMP_RNDU);
    mpfr_setulp (err_diff, D);
    mpfr_add (err_diff, err_diff, t1, GMP_RNDU);
    mpfr_add (err_diff, err_diff, t2, GMP_RNDU);
    mpfr_mul (err_diff, err_diff, res, GMP_RNDU);
    mpfr_mul_2si (err_diff, err_diff, -1, GMP_RNDU);

    mpfr_mul (res, res, D, GMP_RNDN);

    mpfr_mul_2si (err_stat, err_stat, 
		  mpfr_get_exp (res) - mpfr_get_prec (res) + 1, GMP_RNDU);

    mpfr_out_str (stdout, DBASE, 5, err_stat, GMP_RNDN);
    printf (" [E_stat]\n");
    mpfr_out_str (stdout, DBASE, 5, err_eval, GMP_RNDN);
#ifdef DTIS
    printf (" [E_eval DTIS]\n");
#else
    printf (" [E_eval]\n");
#endif
    mpfr_out_str (stdout, DBASE, 5, err_diff, GMP_RNDN);
    printf (" [E_diff]\n");

    mpfr_set (errb, err_diff, GMP_RNDU);
    mpfr_add (errb, errb, err_eval, GMP_RNDU);
    mpfr_add (errb, errb, err_stat, GMP_RNDU);

    mpfr_clear (acc);
    mpfr_clear (fx);
    mpfr_clear (D);
    mpfr_clear (x);
    mpfr_clear (foo);
    mpfr_clear (bar);
    mpfr_clear (apb);
    mpfr_clear (t1);
    mpfr_clear (t2);
    mpfr_clear (err_eval);
    mpfr_clear (err_stat);
    mpfr_clear (err_diff);
    mpfr_clear (curr_deltati);
    mpfr_clear (max_deltati);
}

double ulp_erreur (mpfr_t ret, mpfr_t exact, mpfr_t approche)
{
    mpfr_t diff;
    double r;
    mpfr_init2 (diff, mpfr_get_prec (exact));

    /* BUGFIX: ajout du test somme nulle */

    if ((mpfr_cmp_ui (approche, 0) == 0) && (mpfr_cmp_ui (exact, 0) != 0)) {
	mpfr_set_inf (ret, 1);
	mpfr_clear (diff);
	return 0;
    }
    if (mpfr_sgn (exact) >= 0) {
	if (mpfr_cmp (exact, approche) >= 0)
	    mpfr_sub (diff, exact, approche, GMP_RNDU);
	else
	    mpfr_sub (diff, approche, exact, GMP_RNDU);
    }
    else {
	if (mpfr_cmp (exact, approche) >= 0)
	    mpfr_sub (diff, approche, exact, GMP_RNDD);
	else
	    mpfr_sub (diff, exact, approche, GMP_RNDD);
    }
    mpfr_abs (diff, diff, GMP_RNDU);
    mpfr_mul_2si (diff, diff, mpfr_get_prec (approche) - approche->_mpfr_exp,
		  GMP_RNDN);

    r = mpfr_get_d1 (diff);
    mpfr_swap (diff, ret);
    mpfr_clear (diff);
    return r;
}

/* Compose m times the closed n points NC method on [a, b]
 * M is a bound (absolute value) on the n-th (or n+1-th, depends on the
 * parity of n) derivative of f
 * m1 a bound (absolute value) on the derivative of f.
 * */
void
compose_nc (mpfr_t res, mpfr_t a, mpfr_t b, unsigned int m,
	    unsigned int n, aqfunc_t f, mpfr_t M, mpfr_t m1, mp_prec_t wp)
{
    mpfr_t bmin, bmax;
    mpfr_t acc;
    mpfr_t foo, bar, errb;
    unsigned int step, i;
    mpz_t *coeffs;
    mpz_t d;
    int st;
    mpfr_t method_errb;
    mp_exp_t lost_bits;
    mpfr_t cumul_error;
    mpfr_t nc_error;
    mpfr_t **ncs;
    mpfr_t **fxs;
    assert (m > 0);

    /* Initialize and compute the Bnk */
    coeffs = malloc (((n + 1) / 2) * sizeof (*coeffs));
    for (i = 0; i < (n + 1) / 2; i++)
	mpz_init (coeffs[i]);
    mpz_init (d);

    mpfr_init (method_errb);
    mpfr_init (errb);

    st = cputime ();
    compute_bnk (coeffs, d, n - 1);
    printf ("compute_bnk took %dms\n", cputime () - st);

    mpfr_init2 (bmin, wp);
    mpfr_init2 (bmax, wp);
    mpfr_init2 (foo, wp);
    mpfr_init2 (bar, wp);
    mpfr_init2 (acc, wp);
    mpfr_init2 (cumul_error, wp);
    mpfr_init2 (nc_error, wp);

    /* Initialise tab for nc_gen calls */
    fxs = malloc (((n+1)/2) * sizeof (*fxs));
    for (i = 0; i < (n+1)/2; i++) {
        fxs[i] = malloc (sizeof (mpfr_t));
        mpfr_init2 (*fxs[i], wp);
    }
    /* Initialise tab for nc_gen return values */
    ncs = malloc (m * sizeof (*ncs));
    for (i = 0; i < m; i++) {
        ncs[i] = malloc (sizeof (mpfr_t));
        mpfr_init2 (*ncs[i], wp);
    }

    mpfr_set (bmin, a, GMP_RNDN);
    mpfr_set_ui (acc, 0, GMP_RNDN);
    mpfr_set_ui (cumul_error, 0, GMP_RNDN);

    for (step = 0; step < m; step++) {
	/* Compute new bmax */
	mpfr_mul_ui (foo, a, m - step - 1, GMP_RNDN);
	mpfr_mul_ui (bar, b, step + 1, GMP_RNDN);
	mpfr_add (bmax, foo, bar, GMP_RNDN);
	mpfr_div_ui (bmax, bmax, m, GMP_RNDN);
	/* Compute NC-closed n point on [bmin, bmax] */
	gen_nc_closed (errb, *ncs[step], bmin, bmax, n, coeffs, d, f, m1, wp,
		       fxs);
	mpfr_add (cumul_error, cumul_error, errb, GMP_RNDU);
	if (step == 0)
	    nc_closed_error (method_errb, bmin, bmax, n, M);
	/* Set bmin for next step */
	mpfr_set (bmin, bmax, GMP_RNDN);
#ifndef NDEBUG
	if (mpfr_nan_p (acc)) {
	    printf ("Nan. step = %d.\n", step);
	    abort ();
	}
#endif
    }

    mpfr_sum (res, (mpfr_ptr *) ncs, m, GMP_RNDN);
    /* mpfr_set (res, acc, GMP_RNDN); */
    for (i = 0; i < (n + 1) / 2; i++)
	mpz_clear (coeffs[i]);
    mpz_clear (d);
    free (coeffs);
    mpfr_clear (acc);
    mpfr_clear (bar);
    mpfr_clear (bmax);
    mpfr_clear (bmin);
    mpfr_clear (foo);
    for (i = 0; i < (n+1)/2; i++) {
	mpfr_clear (*fxs[i]);
	free (fxs[i]);
    }
    free (fxs);
    for (i = 0; i < m; i++) {
	mpfr_clear (*ncs[i]);
	free (ncs[i]);
    }
    free (ncs);
    /* Compute a bound on the method error */

    printf ("The method error is bounded by: ");
    mpfr_out_str (stdout, DBASE, 5, method_errb, GMP_RNDN);
    printf (" on each interval\n");
    mpfr_mul_ui (method_errb, method_errb, m, GMP_RNDU);
    mpfr_out_str (stdout, DBASE, 5, method_errb, GMP_RNDN);
    printf (" [accumulated method error]\n");
    mpfr_out_str (stdout, DBASE, 5, cumul_error, GMP_RNDN);
    printf (" [accumulated truncation error]\n");

    mpfr_set (nc_error, cumul_error, GMP_RNDN);
    mpfr_add (nc_error, nc_error, method_errb, GMP_RNDU);
    mpfr_out_str (stdout, DBASE, 5, nc_error, GMP_RNDN);
    printf (" [accumulated TOTAL error]\n");
    if (mpfr_number_p (nc_error) && mpfr_number_p (res))
	lost_bits = mpfr_get_exp (nc_error) - mpfr_get_exp (res) +
		    mpfr_get_prec (res);
    else
	lost_bits = mpfr_get_prec (res);
    printf ("Accurate bits: %li\n", mpfr_get_prec (res) - lost_bits);
    mpfr_clear (method_errb);
    mpfr_clear (nc_error);
    mpfr_clear (cumul_error);
    mpfr_clear (errb);
}

#ifdef TEST2
void sin_wrapper (mpfr_t res, mpfr_t x, mp_prec_t rndm)
{
    mpfr_sin (res, x, rndm);
}

void square_wrapper (mpfr_t res, mpfr_t x, mp_prec_t rndm)
{
    mpfr_mul (res, x, x, rndm);
}

void exp_wrapper (mpfr_t res, mpfr_t x, mp_prec_t rndm)
{
    mpfr_exp (res, x, rndm);
}

void nth_deriv_bound_exp (mpfr_t res, mpfr_t a, mpfr_t b, unsigned int n)
{
    mpfr_exp (res, b, GMP_RNDU);
}

/* t*log(1+t) */
void bailey1 (mpfr_t res, mpfr_t t, mp_prec_t rndm)
{
    mpfr_log1p (res, t, rndm);
    mpfr_mul (res, res, t, rndm);
}

int main (int argc, char *argv[])
{
    unsigned int n, m;
    mp_prec_t prec, wp = 0;

    mpfr_t a, b;
    mpfr_t res;
    mpfr_t M, m1;
    mpfr_t exact_value;
    mpfr_t erreur;
    aqderiv_bound_t *nth_deriv_bound;


    m = 1;
    n = 42;

    while (argc > 1 && argv[1][0] == '-') {
	if (argc >= 3 && strcmp (argv[1], "-p") == 0) {
	    prec = atoi (argv[2]);	/* target precision */
	    mpfr_set_default_prec (prec);
	    argv += 2;
	    argc -= 2;
	}
	else if (argc >= 3 && strcmp (argv[1], "-wp") == 0) {
	    wp = atoi (argv[2]);	/* working precision */
	    argv += 2;
	    argc -= 2;
	}
	else if (argc >= 3 && strcmp (argv[1], "-n") == 0) {
	    n = atoi (argv[2]);	/* Order of NC method */
	    argv += 2;
	    argc -= 2;
	}
	else if (argc >= 3 && strcmp (argv[1], "-m") == 0) {
	    m = atoi (argv[2]);	/* number of intervals */
	    argv += 2;
	    argc -= 2;
	}
	else {
	    fprintf (stderr, "Invalid option: %s\n", argv[1]);
	    exit (1);
	}
    }

    if (wp == 0)
	wp = mpfr_get_default_prec ();

    mpfr_init2 (a, wp);
    mpfr_init2 (b, wp);
    mpfr_init2 (res, wp);
    mpfr_init2 (exact_value, wp);
    mpfr_init2 (erreur, wp);

    mpfr_set_ui (a, LEFT_BOUND, GMP_RNDN);
    mpfr_set_ui (b, RIGHT_BOUND, GMP_RNDN);

    printf ("%d [n]\n%d [m]\n%lu [wp]\n", n, m, wp);

    mpfr_init (M);
    mpfr_init (m1);
#if (FUNCTION == exp_wrapper)
    nth_deriv_bound = nth_deriv_bound_exp;
#endif
    nth_deriv_bound (m1, a, b, 1);
    if (n & 1)
	nth_deriv_bound (M, a, b, n + 1);
    else
	nth_deriv_bound (M, a, b, n);
    compose_nc (res, a, b, m, n, FUNCTION, M, m1, wp);

    printf ("val = ");
    mpfr_out_str (NULL, DBASE, 0, res, GMP_RNDN); printf ("\n");
#if (FUNCTION == exp_wrapper)
    mpfr_set_ui (exact_value, RIGHT_BOUND, GMP_RNDN);
    mpfr_exp (exact_value, exact_value, GMP_RNDN);
    mpfr_set_ui (a, LEFT_BOUND, GMP_RNDN);
    mpfr_exp (a, a, GMP_RNDN);
    mpfr_sub (exact_value, exact_value, a, GMP_RNDN);
    printf ("Exact val: ");
    mpfr_out_str (NULL, DBASE, 0, exact_value, GMP_RNDN); printf ("\n");
    /* derreur = ulp_erreur (erreur, exact_value, res);
    printf ("\nErreur (inexact): %f ulp\n", derreur); */
    mpfr_sub (exact_value, exact_value, res, GMP_RNDN);
    mpfr_abs (exact_value, exact_value, GMP_RNDN);
    mpfr_out_str (NULL, DBASE, 0, exact_value, GMP_RNDN);
    printf (" [measured error]\n");
#endif
    mpfr_clear (a);
    mpfr_clear (b);
    mpfr_clear (erreur);
    mpfr_clear (exact_value);
    mpfr_clear (res);
    mpfr_clear (M);
    mpfr_clear (m1);
    return 0;
}

#endif

#ifdef TEST3

int main (void) {
    mpfr_t x, ulpx;

    mpfr_init2 (x, 4);
    mpfr_init (ulpx);

    mpfr_set_ui (x, 3, GMP_RNDN);
    mpfr_setulp (ulpx, x);
    mpfr_out_str (stdout, DBASE, 0, x, GMP_RNDN);
    printf (" [x]\n");
    mpfr_out_str (stdout, DBASE, 0, ulpx, GMP_RNDN);
    printf (" [ulpx]\n");

}

#endif

#ifdef TEST1

int main (int argc, char *argv[])
{
    unsigned long n = atoi (argv[1]);
    unsigned long k;
    mpz_t *b, d;

    mpz_init (d);
    b = (mpz_t *) malloc ((n / 2 + 1) * sizeof (mpz_t));
    for (k = 0; k <= n / 2; k++)
	mpz_init (b[k]);
    compute_bnk (b, d, n);
    for (k = 0; k <= n; k++) {
	printf ("b[%lu]=", k);
	mpz_out_str (stdout, 10, b[(k <= n / 2) ? k : n - k]);
	printf ("/");
	mpz_out_str (stdout, 10, d);
	printf ("\n");
    }
    for (k = 0; k <= n / 2; k++)
	mpz_clear (b[k]);
    free (b);
    mpz_clear (d);

    return 0;
}

#endif
