/* utils.c -- misc. utility functions.

Copyright 2005 Laurent Fousse <laurent@komite.net>

This file is part of the CRQ Library.

The CRQ Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or (at your
option) any later version.

The CRQ Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with the CRQ Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Place, Fifth Floor, Boston,
MA 02110-1301, USA. */

#include <sys/time.h>
#include <sys/resource.h>

#include "crq-impl.h"

int cputime (void)
{
    struct rusage rus;

    getrusage (0, &rus);
    return rus.ru_utime.tv_sec * 1000 + rus.ru_utime.tv_usec / 1000;
}

void sign_set (char *signmask, int pos, int value) {
    char tmp;
    tmp = ~(1 << pos);
    tmp = *signmask & tmp;
    if (value < 0) {
	tmp |= 1 << pos;
	*signmask = tmp;
    }
    else
	*signmask = tmp;
}

int sign_get (const char signmask, int pos) {
    if (signmask & (1 << pos))
	return -1;
    else
	return 1;
}

int sign_eval (mpz_t c, mpz_t *P, unsigned long k, unsigned long n)
{
    mpz_t v;
    mpz_t t;
    unsigned int i;
    int ret;
    mpz_init (v);
    mpz_init (t);

    mpz_set (v, P[n]);
    i = n;
    do {
	i--;
	mpz_mul (v, v, c);
	mpz_mul_2exp (t, P[i], k * (n - i));
	mpz_add (v, v, t);
    } while (i > 0);
    ret = mpz_sgn (v);
    mpz_clear (t);
    mpz_clear (v);
    return ret;
}

void debug_mpfr_value (mpfr_t foo, char *name)
{
    printf ("%s = ", name);
    mpfr_out_str (stdout, DBASE, 0, foo, GMP_RNDN);
    printf ("\n");
}

void abort_if_nan (mpfr_t foo, char *file, int line)
{
    if (mpfr_nan_p (foo)) {
	printf ("NaN in file %s, line %d.\n", file, line);
	abort ();
    }
}

/* Set ret to ulp(a), to zero if a = 0 */

void mpfr_setulp (mpfr_t ret, mpfr_t a) {
    mp_exp_t exp_a;

    if (mpfr_zero_p (a)) {
	mpfr_set_ui (ret, 0, GMP_RNDN);
	return;
    }
    exp_a = mpfr_get_exp (a);
    mpfr_set_ui_2exp (ret, 1, exp_a - mpfr_get_prec (a), GMP_RNDN);
}

void dump_poly (FILE *out, mpz_t *P, unsigned long d)
{
    unsigned long i;

    for (i = 0; i <= d; i++)
	gmp_fprintf (out, "%Zd\n", P[i]);
}

void load_poly (FILE *in, mpz_t *P, unsigned long d)
{
    unsigned long i;

    for (i = 0; i <= d; i++)
	gmp_fscanf (in, "%Zd\n", P[i]);
}

void dump_interval (FILE *out, interval I)
{
    gmp_fprintf (out, "%Zd %ld\n", I.c, I.k);
}

void load_interval (FILE *in, interval *I)
{
    gmp_fscanf (in, "%Zd %ld\n", I->c, &(I->k));
}
